import inspect
import json
from unittest import mock

from satcfdi.apis.banxico import SieAPIRest
from tests.utils import verify_result


def test_banxico():
    bmx_token = "xxx"

    api = SieAPIRest(
        bmx_token=bmx_token
    )

    value = {
        "bmx": {
            "series": [
                {
                    "idSerie": "SF343410",
                    "titulo": "Tipo de Cambio Cierre de Jornada",
                    "datos": [
                        {
                            "fecha": "11/11/2022",
                            "dato": "19.4920"
                        }
                    ]
                }
            ]
        }
    }
    with mock.patch(f'requests.request') as mk:
        mk.return_value.status_code = 200
        mk.return_value.json = mock.Mock(
            return_value=value
        )

        res = api.datos_oportuno_series_using_get(
            id_series='SF343410'
        )

        assert res == value

        assert mk.called
        mk.call_args.kwargs["headers"]["User-Agent"] = 'this is a test'

        args = json.dumps(mk.call_args.kwargs, indent=2)
        verify = verify_result(data=args, filename=f"{inspect.stack()[0][3]}.json")
        assert verify



from satcfdi.create.cfd.catalogos import Impuesto


def test_catalog():
    _impuestos = {
        "ISR": "001",
        "IVA": "002",
        "IEPS": "003",
    }

    impuesto = "ISR"
    r = _impuestos.get(impuesto, impuesto)
    assert r == '001'

    # v = Impuesto(impuesto)
    # print(v)

    # search enum by name without throwing exception
    v = Impuesto['IVA']
    assert v == Impuesto.IVA
    assert v.value == '002'
    assert v == '002'

    v = Impuesto.get('002', '002')
    assert v == Impuesto.IVA




"""registrofiscal http://www.sat.gob.mx/registrofiscal"""
from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ... import CFDI, XElement, ScalarMap


class CFDIRegistroFiscal(CFDI):
    """
    Complemento para incluir los datos de identificación de los CFDIs generados en Registro Fiscal.
    
    :param folio: Atributo requerido para expresar la relación del CFDI con el Registro Fiscal.
    """
    
    tag = '{http://www.sat.gob.mx/registrofiscal}CFDIRegistroFiscal'
    version = '1.0'
    
    def __init__(
            self,
            folio: str,
    ): 
        super().__init__({
            'Version': self.version,
            'Folio': folio,
        })
        




import json
import os
from datetime import datetime
from decimal import Decimal
from unittest import mock
from satcfdi import CFDI
from satcfdi.create.cancela import cancelacionretencion
from satcfdi.create.cancela.aceptacionrechazo import SolicitudAceptacionRechazo, Folios
from satcfdi.create.cancela.cancelacion import Cancelacion, Folio
from satcfdi.create.cfd import cfdi40
from satcfdi.pacs import Environment, Accept, CancelReason
from satcfdi.pacs.swsapien import SWSapien
from utils import get_signer, verify_result

current_dir = os.path.dirname(__file__)

swsapien = SWSapien(
    token="token",
    user="user",
    password="password",
    environment=Environment.TEST
)


def test_swsapien_test():
    signer = get_signer('xia190128j61', get_csd=True)
    emisor = cfdi40.Emisor(
        rfc=signer.rfc,
        nombre=signer.legal_name,
        regimen_fiscal="601"
    )
    assert emisor["Nombre"] == "XENON INDUSTRIAL ARTICLES"

    invoice = cfdi40.Comprobante(
        emisor=emisor,
        lugar_expedicion="27200",
        fecha=datetime.fromisoformat("2022-09-28T22:40:38"),
        receptor=cfdi40.Receptor(
            rfc="URE180429TM6",
            nombre="UNIVERSIDAD ROBOTICA ESPAÑOLA",
            uso_cfdi="G03",
            domicilio_fiscal_receptor="65000",
            regimen_fiscal_receptor="601"
        ),
        metodo_pago="PPD",
        forma_pago="99",
        serie="T",
        folio="1000",
        conceptos=cfdi40.Concepto(
            cuenta_predial='1234567890',
            clave_prod_serv='10101702',
            cantidad=Decimal('.10'),
            clave_unidad='E48',
            descripcion='SERVICIOS DE RENTA',
            valor_unitario=Decimal("1.00"),
            impuestos=cfdi40.Impuestos(
                traslados=cfdi40.Impuesto.parse('002|Tasa|0.160000'),
                retenciones=[cfdi40.Impuesto.parse('001|Tasa|0.100000'), cfdi40.Impuesto.parse('002|Tasa|0.106667')],
            ),
            _traslados_incluidos=True
        )
    )
    invoice.sign(signer)

    with mock.patch(f'requests.request') as mk:
        mk.return_value.ok = True
        mk.return_value.json = mock.Mock(return_value={
            "status": "success",
            "data": {
                "uuid": '66ea0f5a-4f00-4406-a573-87eb1a34f1fa',
                "cfdi": "dGhpcyBpcyBhIHRlc3QgeG1s"
            }
        })

        res = swsapien.stamp(
            cfdi=invoice,
            accept=Accept.XML
        )

        assert mk.called
        mk.call_args.kwargs["headers"]["User-Agent"] = 'this is a test'

        args = json.dumps(mk.call_args.kwargs, indent=2, default=str, ensure_ascii=False)
        verify = verify_result(data=args, filename="test_stamp.json")
        assert verify


# @pytest.mark.skip(reason="skiping for now")
def test_connectia_cancel():
    cfdi = CFDI(
        {
            "Emisor": {
                "Rfc": "EKU9003173C9"
            },
            "Complemento": {
                "TimbreFiscalDigital": {
                    "UUID": "FE4E71B0-8959-4FB9-8091-F5AC4FB0FEF8"
                }
            }
        }
    )
    cfdi.tag = '{http://www.sat.gob.mx/cfd/3}Comprobante',

    with mock.patch(f'requests.request') as mk:
        mk.return_value.ok = True
        mk.return_value.json = mock.Mock(return_value={
            "data": {
                "acuse": "<?xml version=\"1.0\" encoding=\"utf-8\"?><Acuse xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" Fecha=\"2022-10-27T11:02:56.2475804\" RfcEmisor=\"EKU9003173C9\"><Folios xmlns=\"http://cancelacfd.sat.gob.mx\"><UUID>FE4E71B0-8959-4FB9-8091-F5AC4FB0FEF8</UUID><EstatusUUID>202</EstatusUUID></Folios><Signature Id=\"SelloSAT\" xmlns=\"http://www.w3.org/2000/09/xmldsig#\"><SignedInfo><CanonicalizationMethod Algorithm=\"http://www.w3.org/TR/2001/REC-xml-c14n-20010315\" /><SignatureMethod Algorithm=\"http://www.w3.org/2001/04/xmldsig-more#hmac-sha512\" /><Reference URI=\"\"><Transforms><Transform Algorithm=\"http://www.w3.org/TR/1999/REC-xpath-19991116\"><XPath>not(ancestor-or-self::*[local-name()='Signature'])</XPath></Transform></Transforms><DigestMethod Algorithm=\"http://www.w3.org/2001/04/xmlenc#sha512\" /><DigestValue>rGaqXD3bdRxOv6PBbIi/Oe9JFTJF6hoYtyx63sVTyxByLo0A9+43SNOpyxMYQqGvPowqkiccPDbF1nel2ABSSQ==</DigestValue></Reference></SignedInfo><SignatureValue>LLZEDLTYoSlTJGJRCxkSlS1zqXDNu4kqQoVimYisqQWT9M+uedJYT5PeV2o0an0gcVphqvco7oQb38vz5gykLQ==</SignatureValue><KeyInfo><KeyName>BF66E582888CC845</KeyName><KeyValue><RSAKeyValue><Modulus>n5YsGT0w5Z70ONPbqszhExfJU+KY3Bscftc2jxUn4wxpSjEUhnCuTd88OK5QbDW3Mupoc61jr83lRhUCjchFAmCigpC10rEntTfEU+7qtX8ud/jJJDB1a9lTIB6bhBN//X8IQDjhmHrfKvfen3p7RxLrFoxzWgpwKriuGI5wUlU=</Modulus><Exponent>AQAB</Exponent></RSAKeyValue></KeyValue></KeyInfo></Signature></Acuse>",
                "uuid": {
                    "FE4E71B0-8959-4FB9-8091-F5AC4FB0FEF8": "202"
                }
            },
            "status": "success"
        })

        res = swsapien.cancel(
            cfdi=cfdi,
            reason=CancelReason.COMPROBANTE_EMITIDO_CON_ERRORES_SIN_RELACION
        )

        assert mk.called
        mk.call_args.kwargs["headers"]["User-Agent"] = 'this is a test'

        args = json.dumps(mk.call_args.kwargs, indent=2, default=str, ensure_ascii=False)
        verify = verify_result(data=args, filename="test_cancelation.json")
        assert verify

        assert res.code == "202"


def test_connectia_mass_cancel():
    signer = get_signer('EKU9003173C9', get_csd=True)

    cancelacion = Cancelacion(
        emisor=signer,
        fecha=datetime.fromisoformat("2021-12-26T18:15:28"),
        folios=[
            Folio(
                uuid="fe4e71b0-8959-4fb9-8091-f5ac4fb0fef8",
                motivo="02",
                folio_sustitucion=""
            )
        ]
    )

    with mock.patch(f'requests.request') as mk:
        mk.return_value.ok = True
        mk.return_value.json = mock.Mock(return_value={
            'data': {
                'acuse': '<?xml version="1.0" encoding="utf-8"?><Acuse xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" Fecha="2022-10-27T21:50:49.4572913" RfcEmisor="EKU9003173C9"><Folios xmlns="http://cancelacfd.sat.gob.mx"><UUID>FE4E71B0-8959-4FB9-8091-F5AC4FB0FEF8</UUID><EstatusUUID>202</EstatusUUID></Folios><Signature Id="SelloSAT" xmlns="http://www.w3.org/2000/09/xmldsig#"><SignedInfo><CanonicalizationMethod Algorithm="http://www.w3.org/TR/2001/REC-xml-c14n-20010315" /><SignatureMethod Algorithm="http://www.w3.org/2001/04/xmldsig-more#hmac-sha512" /><Reference URI=""><Transforms><Transform Algorithm="http://www.w3.org/TR/1999/REC-xpath-19991116"><XPath>not(ancestor-or-self::*[local-name()=\'Signature\'])</XPath></Transform></Transforms><DigestMethod Algorithm="http://www.w3.org/2001/04/xmlenc#sha512" /><DigestValue>9z9FlY3o/qJdn62GlDkQYCxD3xQQ/1GQDVpnvJz7/pvJ5hiOkhfY3LGyLfDqbJ7e0u+yOYMJfkYIb/jgwxdPjw==</DigestValue></Reference></SignedInfo><SignatureValue>AvHF37ni19Z889M7YGi2mw94VbT6doP1l4G5c//cGEspaqeuh/0JW2HHGLWbNUgRKPE+GO3OjzWSjt6hEfv61Q==</SignatureValue><KeyInfo><KeyName>BF66E582888CC845</KeyName><KeyValue><RSAKeyValue><Modulus>n5YsGT0w5Z70ONPbqszhExfJU+KY3Bscftc2jxUn4wxpSjEUhnCuTd88OK5QbDW3Mupoc61jr83lRhUCjchFAmCigpC10rEntTfEU+7qtX8ud/jJJDB1a9lTIB6bhBN//X8IQDjhmHrfKvfen3p7RxLrFoxzWgpwKriuGI5wUlU=</Modulus><Exponent>AQAB</Exponent></RSAKeyValue></KeyValue></KeyInfo></Signature></Acuse>',
                'uuid': {'FE4E71B0-8959-4FB9-8091-F5AC4FB0FEF8': '202'}
            },
            'status': 'success'
        })

        res = swsapien.cancel_comprobante(
            cancelacion
        )

        assert mk.called
        mk.call_args.kwargs["headers"]["User-Agent"] = 'this is a test'

        args = json.dumps(mk.call_args.kwargs, indent=2, default=str, ensure_ascii=False)
        verify = verify_result(data=args, filename="test_cancel_xml.json")
        assert verify

        assert res.code == {'FE4E71B0-8959-4FB9-8091-F5AC4FB0FEF8': '202'}


def test_connectia_mass_retencion_cancel():
    signer = get_signer('EKU9003173C9', get_csd=True)

    cancelacion = cancelacionretencion.Cancelacion(
        emisor=signer,
        fecha=datetime.fromisoformat("2021-12-26T18:15:28"),
        folios=[
            Folio(
                uuid="fe4e71b0-8959-4fb9-8091-f5ac4fb0fef8",
                motivo="02",
                folio_sustitucion=""
            )
        ]
    )

    # print(
    #     etree.tostring(cancelacion.to_xml(), pretty_print=True).decode()
    # )

    with mock.patch(f'requests.request') as mk:
        mk.return_value.ok = True
        mk.return_value.json = mock.Mock(return_value={
            'data': {
                'acuse': '<?xml version="1.0" encoding="utf-8"?><Acuse xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" Fecha="2022-10-27T21:50:49.4572913" RfcEmisor="EKU9003173C9"><Folios xmlns="http://cancelacfd.sat.gob.mx"><UUID>FE4E71B0-8959-4FB9-8091-F5AC4FB0FEF8</UUID><EstatusUUID>202</EstatusUUID></Folios><Signature Id="SelloSAT" xmlns="http://www.w3.org/2000/09/xmldsig#"><SignedInfo><CanonicalizationMethod Algorithm="http://www.w3.org/TR/2001/REC-xml-c14n-20010315" /><SignatureMethod Algorithm="http://www.w3.org/2001/04/xmldsig-more#hmac-sha512" /><Reference URI=""><Transforms><Transform Algorithm="http://www.w3.org/TR/1999/REC-xpath-19991116"><XPath>not(ancestor-or-self::*[local-name()=\'Signature\'])</XPath></Transform></Transforms><DigestMethod Algorithm="http://www.w3.org/2001/04/xmlenc#sha512" /><DigestValue>9z9FlY3o/qJdn62GlDkQYCxD3xQQ/1GQDVpnvJz7/pvJ5hiOkhfY3LGyLfDqbJ7e0u+yOYMJfkYIb/jgwxdPjw==</DigestValue></Reference></SignedInfo><SignatureValue>AvHF37ni19Z889M7YGi2mw94VbT6doP1l4G5c//cGEspaqeuh/0JW2HHGLWbNUgRKPE+GO3OjzWSjt6hEfv61Q==</SignatureValue><KeyInfo><KeyName>BF66E582888CC845</KeyName><KeyValue><RSAKeyValue><Modulus>n5YsGT0w5Z70ONPbqszhExfJU+KY3Bscftc2jxUn4wxpSjEUhnCuTd88OK5QbDW3Mupoc61jr83lRhUCjchFAmCigpC10rEntTfEU+7qtX8ud/jJJDB1a9lTIB6bhBN//X8IQDjhmHrfKvfen3p7RxLrFoxzWgpwKriuGI5wUlU=</Modulus><Exponent>AQAB</Exponent></RSAKeyValue></KeyValue></KeyInfo></Signature></Acuse>',
                'uuid': {'FE4E71B0-8959-4FB9-8091-F5AC4FB0FEF8': '202'}
            },
            'status': 'success'
        })

        res = swsapien.cancel_comprobante(
            cancelacion
        )

        assert mk.called
        mk.call_args.kwargs["headers"]["User-Agent"] = 'this is a test'

        args = json.dumps(mk.call_args.kwargs, indent=2, default=str, ensure_ascii=False)
        verify = verify_result(data=args, filename="test_cancel_retencion.json")
        assert verify

        assert res.code == {'FE4E71B0-8959-4FB9-8091-F5AC4FB0FEF8': '202'}


def test_swsapien_relacionados():
    cfdi = CFDI(
        {
            "Emisor": {
                "Rfc": "EKU9003173C9"
            },
            "Complemento": {
                "TimbreFiscalDigital": {
                    "UUID": "362efe1f-c0b4-47dd-9979-91d068a2bc75"
                }
            }
        }
    )
    cfdi.tag = '{http://www.sat.gob.mx/cfd/3}Comprobante',

    with mock.patch(f'requests.request') as mk:
        mk.return_value.ok = True
        mk.return_value.json = mock.Mock(return_value={'codStatus': '2001', 'data': None, 'message': 'No se encontraron CFDI relacionados. ', 'status': 'success'})

        res = swsapien.relations(
            cfdi
        )

        assert mk.called
        mk.call_args.kwargs["headers"]["User-Agent"] = 'this is a test'

        args = json.dumps(mk.call_args.kwargs, indent=2, default=str, ensure_ascii=False)
        verify = verify_result(data=args, filename="test_relacionados.json")
        assert verify

        assert res['codStatus'] == '2001'


def test_swsapien_accept_reject():
    signer = get_signer('EKU9003173C9', get_csd=True)

    request = SolicitudAceptacionRechazo(
        rfc_receptor=signer,
        rfc_pac_envia_solicitud=swsapien.RFC,
        fecha=datetime.fromisoformat("2018-09-20T14:48:09"),
        folios=[
            Folios(
                uuid="FD74D156-B9B0-44A5-9906-E08182E8363E",
                respuesta="Aceptacion"
            )
        ]
    )

    with mock.patch(f'requests.request') as mk:
        mk.return_value.ok = True
        mk.return_value.json = mock.Mock(return_value={
            'data': {
                'acuse': '<?xml version="1.0" encoding="utf-8"?><AcuseAceptacionRechazo xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" RfcReceptor="EKU9003173C9" RfcPac="SPR190613I52" CodEstatus="1000" Fecha="2022-10-28T12:20:33.885126"><Folios Respuesta="Aceptacion" xmlns="http://cancelacfd.sat.gob.mx"><UUID>FD74D156-B9B0-44A5-9906-E08182E8363E</UUID><EstatusUUID>1001</EstatusUUID></Folios><Signature Id="SelloSAT" xmlns="http://www.w3.org/2000/09/xmldsig#"><SignedInfo><CanonicalizationMethod Algorithm="http://www.w3.org/TR/2001/REC-xml-c14n-20010315" /><SignatureMethod Algorithm="http://www.w3.org/2001/04/xmldsig-more#hmac-sha512" /><Reference URI=""><Transforms><Transform Algorithm="http://www.w3.org/TR/1999/REC-xpath-19991116"><XPath>not(ancestor-or-self::*[local-name()=\'Signature\'])</XPath></Transform></Transforms><DigestMethod Algorithm="http://www.w3.org/2001/04/xmlenc#sha512" /><DigestValue>dFlQ/PTBmAAX0I8CBmmU04cr1EJhP6CTbRGuYXeE/4NI6iGi/PRKGDCaM1lxoDihRU8Kcy770QQ+hLNrmWIAmw==</DigestValue></Reference></SignedInfo><SignatureValue>FVpIur/3IyXObIxoit1kc3+/Em2hY9OTStJAxBD8aF+qIplTxOKTJ8fTL6fCjkf9lHGTvshkwPzJUXtYLu6fGw==</SignatureValue><KeyInfo><KeyName>BF66E582888CC845</KeyName><KeyValue><RSAKeyValue><Modulus>n5YsGT0w5Z70ONPbqszhExfJU+KY3Bscftc2jxUn4wxpSjEUhnCuTd88OK5QbDW3Mupoc61jr83lRhUCjchFAmCigpC10rEntTfEU+7qtX8ud/jJJDB1a9lTIB6bhBN//X8IQDjhmHrfKvfen3p7RxLrFoxzWgpwKriuGI5wUlU=</Modulus><Exponent>AQAB</Exponent></RSAKeyValue></KeyValue></KeyInfo></Signature></AcuseAceptacionRechazo>',
                'folios': [{'uuid': 'FD74D156-B9B0-44A5-9906-E08182E8363E', 'estatusUUID': '1001', 'respuesta': 'Aceptacion'}]
            },
            'status': 'success'
        })

        res = swsapien.accept_reject(
            request
        )

        assert mk.called
        mk.call_args.kwargs["headers"]["User-Agent"] = 'this is a test'

        args = json.dumps(mk.call_args.kwargs, indent=2, default=str, ensure_ascii=False)
        verify = verify_result(data=args, filename="test_accept_reject.json")
        assert verify

        assert res.folios == {'FD74D156-B9B0-44A5-9906-E08182E8363E': {'estatusUUID': '1001', 'respuesta': 'Aceptacion'}}


def test_swsapien_pending():
    with mock.patch(f'requests.request') as mk:
        mk.return_value.ok = True
        mk.return_value.json = mock.Mock(return_value={
            'codStatus': '1100',
            'message': 'CA1100 - Se recibío la respuesta de la petición de forma exitosa.',
            'data': {'uuid': ['C8F4B71E-2B2F-439B-8118-D044D8B5F444', '0EC7251E-7CB9-4CFB-A05E-88C0074A89D7', '7FDA7D8C-1B1A-4E24-9AF6-7F7CB3950D51'], 'codEstatus': '1100'},
            'status': 'success'
        })

        res = swsapien.pending("EKU9003173C9")

        assert mk.called
        mk.call_args.kwargs["headers"]["User-Agent"] = 'this is a test'

        args = json.dumps(mk.call_args.kwargs, indent=2, default=str, ensure_ascii=False)
        verify = verify_result(data=args, filename="test_pending.json")
        assert verify

        assert res == ['C8F4B71E-2B2F-439B-8118-D044D8B5F444', '0EC7251E-7CB9-4CFB-A05E-88C0074A89D7', '7FDA7D8C-1B1A-4E24-9AF6-7F7CB3950D51']


def test_swsapien_recover():
    with mock.patch(f'requests.request') as mk:
        mk.return_value.ok = True
        mk.return_value.json = mock.Mock(return_value={
            'data': {'metaData': {'page': 1, 'perPage': 10, 'pageCount': 0, 'totalCount': 0, 'links': {'current': '?page=1&perPage=10'}}, 'records': []}, 'status': 'success'
        })

        res = swsapien.recover(
            document_id="ab21a13a-7385-4493-bec1-52cfb53ec9a7"
        )

        assert mk.called
        mk.call_args.kwargs["headers"]["User-Agent"] = 'this is a test'

        args = json.dumps(mk.call_args.kwargs, indent=2, default=str, ensure_ascii=False)
        verify = verify_result(data=args, filename="recover.json")
        assert verify


def test_swsapien_list_69b():
    pass
    # unable to test



import logging
from enum import Enum

logger = logging.getLogger(__name__)


class Code:
    def __init__(self, code, description):
        self.code = code
        self.description = description

    def __str__(self):
        if self.description is None:
            return self.code
        if isinstance(self.description, list):
            return str(self.code) + " - " + "; ".join(self.description)
        return str(self.code) + " - " + str(self.description)

    def __repr__(self):
        # return '%s.%s(%s)' % (self.__class__.__module__,
        #                       self.__class__.__qualname__,
        #                       f'{repr(self.code)}, {repr(self.description)}')

        return '%s(%s)' % (
            self.__class__.__qualname__,
            f'{repr(self.code)}, {repr(self.description)}'
        )

    def __eq__(self, other):
        if isinstance(other, Code):
            return self.code == other.code
        return self.code == other

    def __hash__(self):
        return self.code.__hash__()





from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence

from ..w3.signature import signature_c14n_sha1
from ... import Signer
from ... import CFDI, XElement, ScalarMap


class Folios(ScalarMap):
    def __init__(
            self,
            respuesta: str,
            uuid: str = None,
    ): 
        """
        
        :param respuesta: 
        :param uuid: 
        """
        
        super().__init__({
            'Respuesta': respuesta,
            'UUID': uuid,
        })
        

class SolicitudAceptacionRechazo(XElement):
    """
    Elemento raíz para realizar la aceptacion o rechazo de una solicitud de cancelacion.
    """
    tag = '{http://cancelacfd.sat.gob.mx}SolicitudAceptacionRechazo'
    
    def __init__(
            self,
            rfc_receptor: Signer = None,
            rfc_pac_envia_solicitud: str = None,
            folios: Folios | Sequence[Folios | dict] = None,
            fecha: datetime = None,
    ):
        """
        Elemento raíz para realizar la aceptacion o rechazo de una solicitud de cancelacion.
        
        :param fecha:
        :param rfc_receptor: 
        :param rfc_pac_envia_solicitud: 
        :param folios: 
        :return: objeto CFDI
        """

        super().__init__({
            'Fecha': fecha or datetime.now(),
            'RfcReceptor': rfc_receptor.rfc,
            'RfcPacEnviaSolicitud': rfc_pac_envia_solicitud,
            'Folios': folios,
        })
        self["_nsmap"] = {
            None: "http://cancelacfd.sat.gob.mx",
            "xsd": "http://www.w3.org/2001/XMLSchema",
            "xsi": "http://www.w3.org/2001/XMLSchema-instance"
        }

        sig = signature_c14n_sha1(
            signer=rfc_receptor,
            element=self.to_xml(),
            nsmap={
                None: 'http://www.w3.org/2000/09/xmldsig#',
                "xsd": "http://www.w3.org/2001/XMLSchema",
                "xsi": "http://www.w3.org/2001/XMLSchema-instance",
            }
        )
        self['Signature'] = sig








import json
import os
from datetime import datetime
from decimal import Decimal

from satcfdi.create.cfd import cfdi40
from satcfdi.pacs.sat import SAT
from tests.utils import get_signer, verify_result, XElementPrettyPrinter

module = 'satcfdi'
current_dir = os.path.dirname(__file__)
current_filename = os.path.splitext(os.path.basename(__file__))[0]
sat = SAT()


def verify_invoice(invoice, path, include_metadata=False):
    if include_metadata:
        verify = verify_result(data=invoice.metadata, filename=f"{path}.txt")
        assert verify

    pp = XElementPrettyPrinter()
    verify = verify_result(data=pp.pformat(invoice), filename=f"{path}.pretty.py")
    assert verify

    verify = verify_result(data=invoice.xml_bytes(pretty_print=True), filename=f"{path}.xml")
    assert verify

    verify = verify_result(data=invoice.html_str(), filename=f"{path}.html")
    assert verify


def test_create_raw():
    signer = get_signer('xiqb891116qe4')

    raw_cfdi = {
        "Fecha": datetime.fromisoformat("2020-01-01T22:40:38"),
        "Version": "4.0",
        "Emisor": {
            'Rfc': signer.rfc,
            'Nombre': signer.legal_name,
            'RegimenFiscal': "601",
        },
        "Receptor": {
            "Rfc": "KIJ0906199R1",
            "Nombre": "KIJ, S.A DE C.V.",
            "DomicilioFiscalReceptor": "59820",
            "RegimenFiscalReceptor": "601",
            "UsoCFDI": "G03"
        },
        "Conceptos": [
            {
                "ClaveProdServ": "10101702",
                "Cantidad": 1,
                "ClaveUnidad": "E48",
                "Descripcion": "SERVICIOS DE FACTURACION",
                "ValorUnitario": Decimal(15390.30),
                "Impuestos": {
                    "Traslados": 'IVA|Tasa|0.160000',
                    "Retenciones": [
                        'ISR|Tasa|0.100000',
                        'IVA|Tasa|0.106667',
                    ]
                },
                "CuentaPredial": "1234567890",
                "_traslados_incluidos": False
            }
        ],
        "LugarExpedicion": "56820",
        "Serie": "A",
        "Folio": "123456",
        "Moneda": "MXN",
    }

    a = cfdi40.Comprobante.__new__(cfdi40.Comprobante)
    a.update(raw_cfdi)
    a.compute()
    a.sign(signer)

    verify_invoice(a, "manual.xml")


def test_create_raw_no_signature():
    signer = get_signer('xiqb891116qe4')

    raw_cfdi = {
        "Fecha": datetime.fromisoformat("2020-01-01T22:40:38"),
        "Version": "4.0",
        "Emisor": {
            'Rfc': signer.rfc,
            'Nombre': signer.legal_name,
            'RegimenFiscal': "601",
        },
        "Receptor": {
            "Rfc": "KIJ0906199R1",
            "Nombre": "KIJ, S.A DE C.V.",
            "DomicilioFiscalReceptor": "59820",
            "RegimenFiscalReceptor": "601",
            "UsoCFDI": "G03"
        },
        "Conceptos": [
            {
                "ClaveProdServ": "10101702",
                "Cantidad": 1,
                "ClaveUnidad": "E48",
                "Descripcion": "SERVICIOS DE FACTURACION",
                "ValorUnitario": Decimal(15390.30),
                "Impuestos": {
                    "Traslados": 'IVA|Tasa|0.160000',
                    "Retenciones": [
                        'ISR|Tasa|0.100000',
                        'IVA|Tasa|0.106667',
                    ]
                },
                "CuentaPredial": "1234567890",
                "_traslados_incluidos": False
            }
        ],
        "LugarExpedicion": "56820",
        "Serie": "A",
        "Folio": "123456",
        "Moneda": "MXN",
        'NoCertificado': '',
        'Certificado': '',
    }

    a = cfdi40.Comprobante.__new__(cfdi40.Comprobante)
    a.update(raw_cfdi)
    a.compute()
    # a.sign(signer)

    verify_invoice(a, "manual_no_signature.xml")


def test_create_no_signature():
    signer = get_signer('xiqb891116qe4')

    invoice = cfdi40.Comprobante(
        emisor=cfdi40.Emisor(
            rfc=signer.rfc,
            nombre=signer.legal_name,
            regimen_fiscal="601",
        ),
        lugar_expedicion="56820",
        fecha=datetime.fromisoformat("2020-01-01T22:40:38"),
        receptor=cfdi40.Receptor(
            rfc='KIJ0906199R1',
            nombre='KIJ, S.A DE C.V.',
            uso_cfdi='G03',
            domicilio_fiscal_receptor="59820",
            regimen_fiscal_receptor="601"
        ),
        metodo_pago='PPD',
        serie="A",
        folio="123456",
        conceptos=[
            cfdi40.Concepto(
                cuenta_predial='1234567890',
                clave_prod_serv='10101702',
                cantidad=Decimal('1.00'),
                clave_unidad='E48',
                descripcion='SERVICIOS DE FACTURACION',
                valor_unitario=Decimal('15390.30'),
                impuestos=cfdi40.Impuestos(
                    traslados='IVA|Tasa|0.160000',
                    retenciones=[
                        'ISR|Tasa|0.100000',
                        'IVA|Tasa|0.106667',
                    ],
                ),
                _traslados_incluidos=False
            )
        ]
    )

    verify_invoice(invoice, "no_signature.xml")




"""PLZ www.sat.gob.mx/esquemas/ContabilidadE/1_1/PolizasPeriodo"""
from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ... import CFDI, XElement, ScalarMap


class OtrMetodoPago(ScalarMap):
    """
    Nodo opcional para relacionar otros métodos de pago de la transacción. Se convierte en requerido cuando la transacción involucra un método de pago diverso a cheque y/o transferencia.
    
    :param met_pago_pol: Atributo requerido para expresar el método de pago de la operación, de acuerdo al catálogo publicado en la página de internet del SAT.
    :param fecha: Atributo requerido, es la fecha de la transacción de otros métodos de pago.
    :param benef: Atributo requerido, nombre de la persona o contribuyente a la cual se realiza éstos métodos de pago.
    :param rfc: Atributo requerido para expresar el RFC relacionado con la transacción. El RFC al que se hace referencia, es el distinto del contribuyente que envía los datos, es decir, el RFC del tercero vinculado.
    :param monto: Atributo requerido para expresar el monto del método de pago soporte de la transacción.
    :param moneda: Atributo opcional para expresar el tipo de moneda utilizado en la transacción, de acuerdo al catálogo publicado en la página de internet del SAT. Este dato sólo se utiliza en el caso de que el tipo de moneda, sea diferente a la moneda nacional (peso). Se convierte en requerido cuando se cuente con la información.
    :param tip_camb: Atributo opcional para expresar el tipo de cambio utilizado de acuerdo al tipo de moneda. Este dato sólo se utiliza en el caso de que el tipo de moneda, sea diferente a la moneda nacional (peso). Se convierte en requerido cuando se cuente con la información.
    """
    
    def __init__(
            self,
            met_pago_pol: str,
            fecha: date,
            benef: str,
            rfc: str,
            monto: Decimal | int,
            moneda: str = None,
            tip_camb: Decimal | int = None,
    ): 
        super().__init__({
            'MetPagoPol': met_pago_pol,
            'Fecha': fecha,
            'Benef': benef,
            'RFC': rfc,
            'Monto': monto,
            'Moneda': moneda,
            'TipCamb': tip_camb,
        })
        

class Transferencia(ScalarMap):
    """
    Nodo opcional para relacionar el detalle de las transferencias bancarias que integran la póliza. Se convierte en requerido cuando exista una salida de recursos que involucre este método de pago por parte del contribuyente que envía los datos. Además se convierte en requerido cuando se realicen transacciones, entre las cuentas propias del contribuyente.
    
    :param banco_ori_nal: Atributo requerido, para expresar el Banco de la cuenta origen de la transferencia, de acuerdo al catálogo publicado en la página de internet del SAT. Se considera banco nacional aquellos bancos de residencia nacional, indistintamente, si el tipo de moneda es nacional o extranjero.
    :param cta_dest: Atributo requerido para expresar el número de cuenta destino, la cual se transfieren los recursos.
    :param banco_dest_nal: Atributo requerido, para expresar el Banco de la cuenta destino de la transferencia, de acuerdo al catálogo publicado en la página de internet del SAT. Se considera banco nacional aquellos bancos de residencia nacional, indistintamente, si el tipo de moneda es nacional o extranjero.
    :param fecha: Atributo requerido, es la fecha de la transferencia
    :param benef: Atributo requerido, nombre del beneficiario de la transferencia.
    :param rfc: Atributo requerido para expresar el RFC relacionado con el movimiento. El RFC al que se hace referencia, es el distinto del contribuyente que envía los datos, es decir, el RFC del tercero vinculado.
    :param monto: Atributo requerido, es el monto transferido
    :param cta_ori: Atributo opcional para expresar el número de cuenta de origen desde la cual se transfieren los recursos. Se convierte en requerido cuando se cuente con la información.
    :param banco_ori_ext: Atributo opcional para expresar el nombre completo del banco origen extranjero. Se convierte en requerido cuando se cuente con la información.
    :param banco_dest_ext: Atributo opcional para expresar el nombre completo del banco destino extranjero. Se convierte en requerido cuando se cuente con la información.
    :param moneda: Atributo opcional para expresar el tipo de moneda utilizado en la transacción, de acuerdo al catálogo publicado en la página de internet del SAT. Este dato sólo se utiliza en el caso de que el tipo de moneda, sea diferente a la moneda nacional (peso). Se convierte en requerido cuando se cuente con la información.
    :param tip_camb: Atributo opcional para expresar el tipo de cambio utilizado de acuerdo al tipo de moneda. Este dato sólo se utiliza en el caso de que el tipo de moneda, sea diferente a la moneda nacional (peso). Se convierte en requerido cuando se cuente con la información.
    """
    
    def __init__(
            self,
            banco_ori_nal: str,
            cta_dest: str,
            banco_dest_nal: str,
            fecha: date,
            benef: str,
            rfc: str,
            monto: Decimal | int,
            cta_ori: str = None,
            banco_ori_ext: str = None,
            banco_dest_ext: str = None,
            moneda: str = None,
            tip_camb: Decimal | int = None,
    ): 
        super().__init__({
            'BancoOriNal': banco_ori_nal,
            'CtaDest': cta_dest,
            'BancoDestNal': banco_dest_nal,
            'Fecha': fecha,
            'Benef': benef,
            'RFC': rfc,
            'Monto': monto,
            'CtaOri': cta_ori,
            'BancoOriExt': banco_ori_ext,
            'BancoDestExt': banco_dest_ext,
            'Moneda': moneda,
            'TipCamb': tip_camb,
        })
        

class Cheque(ScalarMap):
    """
    Nodo opcional para relacionar el detalle de los cheques que integran la póliza. Se convierte en requerido cuando exista una salida de recursos, que involucre este método de pago de la obligación contraída por parte del contribuyente que envía los datos
    
    :param num: Atributo requerido para expresar el número del cheque emitido
    :param ban_emis_nal: Atributo requerido, para expresar el Banco nacional emisor del cheque, de acuerdo al catálogo publicado en la página de internet del SAT. Se consideran banco nacional aquellos bancos de residencia nacional, indistintamente, si el tipo de moneda es nacional o extranjero.
    :param cta_ori: Atributo requerido para expresar el número de cuenta bancaria del origen de los recursos.
    :param fecha: Atributo requerido, es la fecha del cheque
    :param benef: Atributo requerido, nombre del beneficiario del cheque
    :param rfc: Atributo requerido para expresar el RFC relacionado con el movimiento. El RFC al que se hace referencia, es el distinto del contribuyente que envía los datos, es decir, el RFC del tercero vinculado.
    :param monto: Atributo requerido, es el monto del cheque emitido
    :param ban_emis_ext: Atributo opcional para expresar el nombre completo del Banco extranjero emisor del cheque. Se convierte en requerido cuando se cuente con la información.
    :param moneda: Atributo opcional para expresar el tipo de moneda utilizado en la transacción, de acuerdo al catálogo publicado en la página de internet del SAT. Este dato sólo se utiliza en el caso de que el tipo de moneda, sea diferente a la moneda nacional (peso). Se convierte en requerido cuando se cuente con la información.
    :param tip_camb: Atributo opcional para expresar el tipo de cambio utilizado de acuerdo al tipo de moneda. Este dato sólo se utiliza en el caso de que el tipo de moneda, sea diferente a la moneda nacional (peso). Se convierte en requerido cuando se cuente con la información.
    """
    
    def __init__(
            self,
            num: str,
            ban_emis_nal: str,
            cta_ori: str,
            fecha: date,
            benef: str,
            rfc: str,
            monto: Decimal | int,
            ban_emis_ext: str = None,
            moneda: str = None,
            tip_camb: Decimal | int = None,
    ): 
        super().__init__({
            'Num': num,
            'BanEmisNal': ban_emis_nal,
            'CtaOri': cta_ori,
            'Fecha': fecha,
            'Benef': benef,
            'RFC': rfc,
            'Monto': monto,
            'BanEmisExt': ban_emis_ext,
            'Moneda': moneda,
            'TipCamb': tip_camb,
        })
        

class CompExt(ScalarMap):
    """
    Nodo opcional para relacionar el detalle de los comprobantes de origen extranjero relacionados con la transacción. Se considera que se debe identificar, el soporte documental, tanto en la provisión, como en el pago y/o cobro de cada una de las cuentas y subcuentas que se vean afectadas. Se convierte en requerido cuando se cuente con la información.
    
    :param num_fact_ext: Atributo requerido para expresar la clave numérico o alfanumérico del comprobante de origen extranjero que soporte la operación
    :param monto_total: Atributo requerido para expresar el monto total del comprobante de origen extranjero que soporte la operación.
    :param tax_id: Atributo opcional que sirve para expresar el Identificador del contribuyente extranjero. Se convierte en requerido cuando se cuente con la información
    :param moneda: Atributo opcional para expresar el tipo de moneda utilizado en la transacción, de acuerdo al catálogo publicado en la página de internet del SAT. Este dato sólo se utiliza en el caso de que el tipo de moneda, sea diferente a la moneda nacional (peso). Se convierte en requerido cuando se cuente con la información.
    :param tip_camb: Atributo opcional para expresar el tipo de cambio utilizado de acuerdo al tipo de moneda. Este dato sólo se utiliza en el caso de que el tipo de moneda, sea diferente a la moneda nacional (peso). Se convierte en requerido cuando se cuente con la información.
    """
    
    def __init__(
            self,
            num_fact_ext: str,
            monto_total: Decimal | int,
            tax_id: str = None,
            moneda: str = None,
            tip_camb: Decimal | int = None,
    ): 
        super().__init__({
            'NumFactExt': num_fact_ext,
            'MontoTotal': monto_total,
            'TaxID': tax_id,
            'Moneda': moneda,
            'TipCamb': tip_camb,
        })
        

class CompNalOtr(ScalarMap):
    """
    Nodo opcional para relacionar el detalle de los comprobantes de origen nacional relacionados con la transacción, diferente a CFDI, es decir, CFD y/o CBB. Se considera que se debe identificar, el soporte documental, tanto en la provisión, como en el pago y/o cobro de cada una de las cuentas y subcuentas que se vean afectadas. Se convierte en requerido cuando se cuente con la información.
    
    :param cfd_cbb_num_fol: Atributo requerido para expresar el número de folio del comprobante CFD_CBB que soporte la transacción.
    :param rfc: Atributo requerido para expresar el RFC relacionado con el movimiento o transacción. El RFC al que se hace referencia, es el distinto del contribuyente que envía los datos, es decir, el RFC del tercero vinculado.
    :param monto_total: Atributo requerido para expresar el monto total del CFD y/o CBB que soporte la transacción. (Incluye IVA en su caso)
    :param cfd_cbb_serie: Atributo opcional para expresar la serie del comprobante CFD_CBB que soporte la transacción.
    :param moneda: Atributo opcional para expresar el tipo de moneda utilizado en la transacción, de acuerdo al catálogo publicado en la página de internet del SAT. Este dato sólo se utiliza en el caso de que el tipo de moneda, sea diferente a la moneda nacional (peso). Se convierte en requerido cuando se cuente con la información.
    :param tip_camb: Atributo opcional para expresar el tipo de cambio utilizado de acuerdo al tipo de moneda. Este dato sólo se utiliza en el caso de que el tipo de moneda, sea diferente a la moneda nacional (peso). Se convierte en requerido cuando se cuente con la información.
    """
    
    def __init__(
            self,
            cfd_cbb_num_fol: int,
            rfc: str,
            monto_total: Decimal | int,
            cfd_cbb_serie: str = None,
            moneda: str = None,
            tip_camb: Decimal | int = None,
    ): 
        super().__init__({
            'CFD_CBB_NumFol': cfd_cbb_num_fol,
            'RFC': rfc,
            'MontoTotal': monto_total,
            'CFD_CBB_Serie': cfd_cbb_serie,
            'Moneda': moneda,
            'TipCamb': tip_camb,
        })
        

class CompNal(ScalarMap):
    """
    Nodo opcional para relacionar el detalle de los comprobantes de origen nacional relacionados con la transacción. Se considera que se debe identificar, el soporte documental, tanto en la provisión, como en el pago y/o cobro de cada una de las cuentas y subcuentas que se vean afectadas. Se convierte en requerido cuando se cuente con la información.
    
    :param uuid_cfdi: Atributo requerido para expresar la clave UUID del CFDI soporte de la operación. (36 caracteres)
    :param rfc: Atributo requerido para expresar el RFC relacionado con el movimiento o transacción. El RFC al que se hace referencia, es el distinto del contribuyente que envía los datos, es decir, el RFC del tercero vinculado.
    :param monto_total: Atributo requerido para expresar el monto total del CFDI que soporte la transacción. (Incluye IVA en su caso)
    :param moneda: Atributo opcional para expresar el tipo de moneda utilizado en la transacción, de acuerdo al catálogo publicado en la página de internet del SAT. Este dato sólo se utiliza en el caso de que el tipo de moneda, sea diferente a la moneda nacional (peso). Se convierte en requerido cuando se cuente con la información.
    :param tip_camb: Atributo opcional para expresar el tipo de cambio utilizado de acuerdo al tipo de moneda. Este dato sólo se utiliza en el caso de que el tipo de moneda, sea diferente a la moneda nacional (peso). Se convierte en requerido cuando se cuente con la información.
    """
    
    def __init__(
            self,
            uuid_cfdi: str,
            rfc: str,
            monto_total: Decimal | int,
            moneda: str = None,
            tip_camb: Decimal | int = None,
    ): 
        super().__init__({
            'UUID_CFDI': uuid_cfdi,
            'RFC': rfc,
            'MontoTotal': monto_total,
            'Moneda': moneda,
            'TipCamb': tip_camb,
        })
        

class Transaccion(ScalarMap):
    """
    Nodo obligatorio para relacionar el detalle de cada transacción dentro de la póliza
    
    :param num_cta: Atributo requerido para expresar la clave con que se distingue la cuenta o subcuenta que se afecta por la transacción.
    :param des_cta: Atributo requerido para expresar el nombre de la cuenta o subcuenta que se afecta por la transacción.
    :param concepto: Atributo requerido para expresar el concepto de la transacción
    :param debe: Atributo requerido para expresar el monto del cargo a la cuenta o subcuenta que se afecta en la transacción. En caso de no existir dato, colocar cero (0)
    :param haber: Atributo requerido para expresar el monto del abono a la cuenta o subcuenta que se afecta en la transacción. En caso de no existir dato, colocar cero (0)
    :param comp_nal: Nodo opcional para relacionar el detalle de los comprobantes de origen nacional relacionados con la transacción. Se considera que se debe identificar, el soporte documental, tanto en la provisión, como en el pago y/o cobro de cada una de las cuentas y subcuentas que se vean afectadas. Se convierte en requerido cuando se cuente con la información.
    :param comp_nal_otr: Nodo opcional para relacionar el detalle de los comprobantes de origen nacional relacionados con la transacción, diferente a CFDI, es decir, CFD y/o CBB. Se considera que se debe identificar, el soporte documental, tanto en la provisión, como en el pago y/o cobro de cada una de las cuentas y subcuentas que se vean afectadas. Se convierte en requerido cuando se cuente con la información.
    :param comp_ext: Nodo opcional para relacionar el detalle de los comprobantes de origen extranjero relacionados con la transacción. Se considera que se debe identificar, el soporte documental, tanto en la provisión, como en el pago y/o cobro de cada una de las cuentas y subcuentas que se vean afectadas. Se convierte en requerido cuando se cuente con la información.
    :param cheque: Nodo opcional para relacionar el detalle de los cheques que integran la póliza. Se convierte en requerido cuando exista una salida de recursos, que involucre este método de pago de la obligación contraída por parte del contribuyente que envía los datos
    :param transferencia: Nodo opcional para relacionar el detalle de las transferencias bancarias que integran la póliza. Se convierte en requerido cuando exista una salida de recursos que involucre este método de pago por parte del contribuyente que envía los datos. Además se convierte en requerido cuando se realicen transacciones, entre las cuentas propias del contribuyente.
    :param otr_metodo_pago: Nodo opcional para relacionar otros métodos de pago de la transacción. Se convierte en requerido cuando la transacción involucra un método de pago diverso a cheque y/o transferencia.
    """
    
    def __init__(
            self,
            num_cta: str,
            des_cta: str,
            concepto: str,
            debe: Decimal | int,
            haber: Decimal | int,
            comp_nal: CompNal | dict | Sequence[CompNal | dict] = None,
            comp_nal_otr: CompNalOtr | dict | Sequence[CompNalOtr | dict] = None,
            comp_ext: CompExt | dict | Sequence[CompExt | dict] = None,
            cheque: Cheque | dict | Sequence[Cheque | dict] = None,
            transferencia: Transferencia | dict | Sequence[Transferencia | dict] = None,
            otr_metodo_pago: OtrMetodoPago | dict | Sequence[OtrMetodoPago | dict] = None,
    ): 
        super().__init__({
            'NumCta': num_cta,
            'DesCta': des_cta,
            'Concepto': concepto,
            'Debe': debe,
            'Haber': haber,
            'CompNal': comp_nal,
            'CompNalOtr': comp_nal_otr,
            'CompExt': comp_ext,
            'Cheque': cheque,
            'Transferencia': transferencia,
            'OtrMetodoPago': otr_metodo_pago,
        })
        

class Poliza(ScalarMap):
    """
    Nodo obligatorio para relacionar el detalle de cada transacción dentro de la póliza.
    
    :param num_un_iden_pol: Atributo requerido para expresar el número único de identificación de la póliza. El campo deberá contener la clave o nombre utilizado por el contribuyente para diferenciar, el tipo de póliza y el número correspondiente. En un mes ordinario no debe repetirse un mismo número de póliza con la clave o nombre asignado por el contribuyente.
    :param fecha: Atributo requerido para expresar la fecha de registro de la póliza
    :param concepto: Atributo requerido para expresar el concepto de la operación
    :param transaccion: Nodo obligatorio para relacionar el detalle de cada transacción dentro de la póliza
    """
    
    def __init__(
            self,
            num_un_iden_pol: str,
            fecha: date,
            concepto: str,
            transaccion: Transaccion | dict | Sequence[Transaccion | dict],
    ): 
        super().__init__({
            'NumUnIdenPol': num_un_iden_pol,
            'Fecha': fecha,
            'Concepto': concepto,
            'Transaccion': transaccion,
        })
        

class Polizas(CFDI):
    """
    Estándar de pólizas del periodo que se entrega como parte de la contabilidad electrónica.
    
    :param rfc: Atributo requerido para expresar el RFC del contribuyente que envía los datos
    :param mes: Atributo requerido para expresar el mes al que corresponde la póliza
    :param anio: Atributo requerido para expresar el año al que corresponde la póliza
    :param tipo_solicitud: Atributo requerido para expresar el tipo de solicitud de la póliza ( AF - Acto de Fiscalización; FC - Fiscalización Compulsa; DE - Devolución; CO - Compensación )
    :param poliza: Nodo obligatorio para relacionar el detalle de cada transacción dentro de la póliza.
    :param num_orden: Atributo opcional para expresar el número de orden asignado al acto de fiscalización al que hace referencia la solicitud de la póliza. Requerido para tipo de solicitud = AF y FC. Se convierte en requerido cuando se cuente con la información.
    :param num_tramite: Atributo opcional para expresar el número de trámite asignado a la solicitud de devolución o compensación al que hace referencia la solicitud de la póliza. Requerido para tipo de solicitud = DE o CO. Se convierte en requerido cuando se cuente con la información.
    :param sello: Atributo opcional para contener el sello digital del archivo de contabilidad electrónica. El sello deberá ser expresado cómo una cadena de texto en formato Base 64
    :param no_certificado: Atributo opcional para expresar el número de serie del certificado de sello digital que ampara el archivo de contabilidad electrónica, de acuerdo al acuse correspondiente a 20 posiciones otorgado por el sistema del SAT.
    :param certificado: Atributo opcional que sirve para expresar el certificado de sello digital que ampara al archivo de contabilidad electrónica como texto, en formato base 64.
    """
    
    tag = '{www.sat.gob.mx/esquemas/ContabilidadE/1_1/PolizasPeriodo}Polizas'
    version = '1.1'
    
    def __init__(
            self,
            rfc: str,
            mes: str,
            anio: int,
            tipo_solicitud: str,
            poliza: Poliza | dict | Sequence[Poliza | dict],
            num_orden: str = None,
            num_tramite: str = None,
            sello: str = None,
            no_certificado: str = None,
            certificado: str = None,
    ): 
        super().__init__({
            'Version': self.version,
            'RFC': rfc,
            'Mes': mes,
            'Anio': anio,
            'TipoSolicitud': tipo_solicitud,
            'Poliza': poliza,
            'NumOrden': num_orden,
            'NumTramite': num_tramite,
            'Sello': sello,
            'NoCertificado': no_certificado,
            'Certificado': certificado,
        })
        




from ..models import Signer, Certificate


# class Issuer:
#     """
#     Instancia generica de Emisor para usar en la creacion y sellado de comprobantes
#     """
#
#     def __init__(self, tax_system: str, signer: Signer | Certificate = None, legal_name: str = None, rfc: str = None, certificate_number: str = None):
#         """
#         Issuer to issue and stamp CFDIs
#
#         :param signer: Llave y Certificado del Emisor
#         :param legal_name: Legal Name (Razon Social)
#         :param tax_system: Atributo requerido para incorporar la clave del régimen del contribuyente emisor al que aplicará el efecto fiscal de este comprobante.
#         """
#         self.signer = signer if isinstance(signer, Signer) else None
#         self.certificate = signer
#         self.rfc = rfc or signer.rfc
#         self.legal_name = legal_name or signer.legal_name
#         self.tax_system = tax_system
#         self.certificate_number = certificate_number or (signer.certificate_number if signer else "")



from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from uuid import UUID

from .. import CFDI, XElement
from ..create.catalogos import EstadoComprobante
from ..create.cfd.catalogos import MetodoPago, TipoDeComprobante, TipoRelacion
from ..create.compute import make_impuestos_dr_parcial, rounder, group_impuestos, encode_impuesto, calculate_partial


class SatCFDI(CFDI):
    """
    SatCFDI is an extension of a CFDI to represent a CFDI that has been sent to SAT
    """

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, *args, **kwargs)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.relations = []  # type: list[Relation]
        self.payments = []  # type: list[Payment]

    @property
    def uuid(self):
        return UUID(self["Complemento"]["TimbreFiscalDigital"]["UUID"])

    @property
    def name(self):
        return self.get("Serie", "") + self.get("Folio", "")

    @property
    def saldo_pendiente(self) -> Decimal | None:
        if self["TipoDeComprobante"] == TipoDeComprobante.INGRESO:
            # Nota de crédito de los documentos relacionados
            credit_notes = sum(
                c.comprobante["Total"]
                for c in self.relations
                if c.cfdi_relacionados["TipoRelacion"] == TipoRelacion.NOTA_DE_CREDITO_DE_LOS_DOCUMENTOS_RELACIONADOS
                and c.comprobante['TipoDeComprobante'] == TipoDeComprobante.EGRESO
                and c.comprobante.estatus == EstadoComprobante.VIGENTE
            )
            insoluto = min(
                (c.docto_relacionado['ImpSaldoInsoluto']
                 for c in self.payments
                 if c.comprobante.estatus == EstadoComprobante.VIGENTE),
                default=None
            )
            if insoluto is not None:
                return insoluto - credit_notes

            insoluto = self["Total"] - credit_notes
            if self["MetodoPago"] == MetodoPago.PAGO_EN_PARCIALIDADES_O_DIFERIDO:
                return insoluto
            if self["MetodoPago"] == MetodoPago.PAGO_EN_UNA_SOLA_EXHIBICION:
                return Decimal(0)

        return None

    @property
    def ultima_num_parcialidad(self) -> int:
        return max((c.docto_relacionado['NumParcialidad'] for c in self.payments), default=0)

    def consulta_estado(self) -> dict:
        raise NotImplementedError()

    @property
    def estatus(self) -> EstadoComprobante:
        raise NotImplementedError()

    @property
    def fecha_cancelacion(self) -> datetime | None:
        raise NotImplementedError()


@dataclass(slots=True, init=True)
class Relation:
    cfdi_relacionados: XElement
    comprobante: SatCFDI


@dataclass(slots=True, init=True)
class Payment:
    comprobante: SatCFDI
    pago: XElement = None
    docto_relacionado: XElement = None


@dataclass
class PaymentsDetails(Payment):
    comprobante_pagado: SatCFDI = None

    def __post_init__(self):
        if self.pago:
            self.impuestos = self.docto_relacionado.get('ImpuestosDR')
            if self.impuestos is None:
                self.impuestos = make_impuestos_dr_parcial(
                    conceptos=self.comprobante_pagado['Conceptos'],
                    imp_saldo_ant=self.docto_relacionado['ImpSaldoAnt'],
                    imp_pagado=self.docto_relacionado['ImpPagado'],
                    total=self.comprobante_pagado['Total'],
                    rnd_fn=rounder(self.comprobante_pagado['Moneda'])
                )

            self.impuestos = group_impuestos([{
                "ImpuestosDR": self.impuestos
            }], pfx="DR", ofx="")

            for imp, imps in self.impuestos.items():
                self.impuestos[imp] = {
                    encode_impuesto(
                        impuesto=v['Impuesto'],
                        tipo_factor=v.get("TipoFactor"),
                        tasa_cuota=v.get('TasaOCuota')
                    ): v
                    for v in imps
                }

            def calc_parcial(field):
                return calculate_partial(
                    value=self.comprobante_pagado.get(field),
                    imp_saldo_ant=self.docto_relacionado['ImpSaldoAnt'],
                    imp_pagado=self.docto_relacionado["ImpPagado"],
                    total=self.comprobante_pagado["Total"],
                    rnd_fn=rounder(self.comprobante_pagado['Moneda'])
                )

            self.sub_total = calc_parcial("SubTotal")
            self.descuento = calc_parcial("Descuento")
            self.total = self.docto_relacionado["ImpPagado"]

        else:
            self.impuestos = self.comprobante.get("Impuestos", {})
            self.sub_total = self.comprobante["SubTotal"]
            self.descuento = self.comprobante.get("Descuento")
            self.total = self.comprobante["Total"]



from datetime import datetime

from satcfdi.create.cancela import cancelacion, cancelacionretencion
from tests.utils import verify_result, get_signer, XElementPrettyPrinter


def verify_invoice(invoice, path):
    pp = XElementPrettyPrinter()
    verify = verify_result(data=pp.pformat(invoice), filename=f"{path}.py")
    assert verify

    verify = verify_result(data=invoice.xml_bytes(pretty_print=True), filename=f"{path}.xml")
    assert verify


def test_create_cancelacion():
    signer = get_signer('emisor2021')

    can = cancelacion.Cancelacion(
        emisor=signer,
        fecha=datetime.fromisoformat("2022-01-29T15:45:12"),
        folios=[
            cancelacion.Folio(
                uuid="E3BA5811-FD07-4B9C-84B4-B146D7560575",
                motivo="01",
                folio_sustitucion="20D816A0-BA64-4DF4-BE87-AA55E62F992F"
            ),
            cancelacion.Folio(
                uuid="29880559-5057-4871-9572-A222FD10BD97",
                motivo="02",
            )
        ]
    )

    verify_invoice(can, "cancelacion")


def test_create_cancelacion_retencion():
    signer = get_signer('emisor2021')

    can = cancelacionretencion.Cancelacion(
        emisor=signer,
        fecha=datetime.fromisoformat("2022-01-29T15:45:12"),
        folios=[
            cancelacionretencion.Folio(
                uuid="E3BA5811-FD07-4B9C-84B4-B146D7560575",
                motivo="01",
                folio_sustitucion="20D816A0-BA64-4DF4-BE87-AA55E62F992F"
            ),
            cancelacionretencion.Folio(
                uuid="29880559-5057-4871-9572-A222FD10BD97",
                motivo="02",
            )
        ]
    )

    verify_invoice(can, "cancelacionretencion")



import inspect
import os
from datetime import date

import pytest

from satcfdi import CFDI, SchemaValidationError
from satcfdi.create.pld import ari
from tests.utils import verify_result, XElementPrettyPrinter

module = 'satcfdi'
current_dir = os.path.dirname(__file__)


def verify_invoice(invoice, path):
    pp = XElementPrettyPrinter()
    verify = verify_result(data=pp.pformat(invoice), filename=f"{path}.pretty.py")
    assert verify

    verify = verify_result(data=invoice.xml_bytes(pretty_print=True), filename=f"{path}.xml")
    assert verify

    verify = verify_result(data=invoice.html_str(), filename=f"{path}.html")
    assert verify


def test_ejemplos():
    pld_ejemplos_dir = os.path.join(current_dir, 'pdl', 'ejemplos')

    for file in os.listdir(pld_ejemplos_dir):
        ejemplo = CFDI.from_file(os.path.join(pld_ejemplos_dir, file))

        verify_invoice(ejemplo, path=f"ejemplos/{file}")


def test_arrendamiento_de_inmuebles_zeros():
    ari_file = os.path.join(current_dir, 'pdl', 'ejemplo_ari_zeros.xml')
    report_f = CFDI.from_file(ari_file)

    report_a = ari.Archivo(
        informe=ari.InformeType(
            mes_reportado='202210',
            sujeto_obligado=ari.SujetoObligadoType(
                clave_sujeto_obligado='OGA751212G56',
                clave_actividad='ARI'
            )
        )
    )

    report_a = report_a.process()
    assert report_a == report_f

    verify = verify_result(data=report_a.xml_bytes(pretty_print=True, include_schema_location=True), filename=f"o_{inspect.stack()[0][3]}.xml")
    assert verify


def test_arrendamiento_de_inmuebles_zeros_invalid():
    with pytest.raises(SchemaValidationError) as excinfo:
        report_a = ari.Archivo(
            informe=ari.InformeType(
                mes_reportado='20221043',  # Mes invalido
                sujeto_obligado=ari.SujetoObligadoType(
                    clave_sujeto_obligado='OGA751212G56',
                    clave_actividad='ARI'
                )
            )
        ).process(validate=True)

    verify = verify_result(data=str(excinfo.value.error_log), filename=f"o_{inspect.stack()[0][3]}.txt")
    assert verify


def test_arrendamiento_de_inmuebles():
    ari_file = os.path.join(current_dir, 'pdl', 'ejemplo_ari.xml')
    report_f = CFDI.from_file(ari_file)

    report_a = ari.Archivo(
        informe=ari.InformeType(
            mes_reportado='201407',
            sujeto_obligado=ari.SujetoObligadoType(
                clave_sujeto_obligado='OGA751212G56',
                clave_actividad='ARI'
            ),
            aviso=ari.AvisoType(
                referencia_aviso='REF15454FG454',
                prioridad='1',
                alerta=ari.AlertaType(
                    tipo_alerta='100'
                ),
                persona_aviso=ari.PersonaAvisoType(
                    tipo_persona=ari.TipoPersonaType(
                        persona_fisica=ari.PersonaFisicaType(
                            nombre='NEPOMUCENO',
                            apellido_paterno='ALMONTE',
                            apellido_materno='JUAREZ',
                            fecha_nacimiento=date(1956, 8, 16),
                            pais_nacionalidad='TG',
                            actividad_economica='3130100'
                        )
                    ),
                    tipo_domicilio=ari.TipoDomicilioType(
                        extranjero=ari.ExtranjeroType(
                            pais='TG',
                            estado_provincia='TOGUILLITA',
                            ciudad_poblacion='TOGUIS',
                            colonia='NA',
                            calle='TOGA TOGA',
                            numero_exterior='45',
                            codigo_postal='12448'
                        )
                    ),
                ),
                detalle_operaciones=ari.DetalleOperacionesType(
                    datos_operacion=ari.DatosOperacionType(
                        fecha_operacion=date(2014, 7, 1),
                        tipo_operacion='1501',
                        caracteristicas=ari.CaracteristicasType(
                            fecha_inicio=date(2014, 1, 1),
                            fecha_termino=date(2015, 1, 1),
                            tipo_inmueble='3',
                            valor_referencia='356825.12',
                            colonia='6920',
                            calle='SAN SIMON TOLNAHUAC',
                            numero_exterior='VIOLANTE',
                            numero_interior='45',
                            codigo_postal='01058',
                            folio_real='BG544-FRR-456B-FRR'
                        ),
                        datos_liquidacion=ari.DatosLiquidacionType(
                            fecha_pago=date(2014, 7, 1),
                            forma_pago='4',
                            instrumento_monetario='4',
                            moneda='2',
                            monto_operacion='757897.55'
                        )
                    )
                )
            )
        )
    ).process()

    assert report_a == report_f

    verify_invoice(report_a, path=f"o_{inspect.stack()[0][3]}")



PAISES = {
    'AD': 'Principado de Andorra',
    'AE': 'Emiratos Arabes Unidos',
    'AF': 'Afganistán',
    'AG': 'Antigua y Bermuda',
    'AI': 'Isla Anguilla',
    'AL': 'República de Albania',
    'AN': 'Antillas Neerlandesas',
    'AO': 'República de Angola',
    'AQ': 'Antártica',
    'AR': 'Argentina',
    'AS': 'Samoa Americana',
    'AT': 'Austria',
    'AU': 'Australia',
    'AW': 'Aruba',
    'AX': 'Ascensión',
    'AZ': 'Islas Azores',
    'BB': 'Barbados',
    'BD': 'Bangladesh',
    'BE': 'Bélgica',
    'BF': 'Burkina Faso',
    'BG': 'Bulgaria',
    'BH': 'Estado de Bahrein',
    'BI': 'Burundi',
    'BJ': 'Benin',
    'BL': 'Belice',
    'BM': 'Bermudas',
    'BN': 'Brunei Darussalam',
    'BO': 'Bolivia',
    'BR': 'Brasil',
    'BS': 'Commonwealth de las Bahamas',
    'BT': 'Buthan',
    'BU': 'Burma',
    'BV': 'Isla Bouvet',
    'BW': 'Botswana',
    'BY': 'Bielorrusia',
    'CA': 'Canadá',
    'CC': 'Isla de Cocos o Kelling',
    'CD': 'Islas Canarias',
    'CE': 'Isla de Christmas',
    'CF': 'República Centro Africana',
    'CG': 'Congo',
    'CH': 'Suiza',
    'CI': 'Costa de Marfil',
    'CK': 'Islas Cook',
    'CL': 'Chile',
    'CM': 'Camerún',
    'CN': 'China',
    'CO': 'Colombia',
    'CP': "Campione D'Italia",
    'CR': 'República de Costa Rica',
    'CS': 'República Checa y República Eslovaca',
    'CU': 'Cuba',
    'CV': 'República de Cabo Verde',
    'CX': 'Isla de Navidad',
    'CY': 'República de Chipre',
    'DD': 'Alemania',
    'DJ': 'República de Djibouti',
    'DK': 'Dinamarca',
    'DM': 'República Dominicana',
    'DN': 'Commonwealth de Dominica',
    'DZ': 'Argelia',
    'EC': 'Ecuador',
    'EG': 'Egipto',
    'EH': 'Sahara del Oeste',
    'EO': 'Estado Independiente de Samoa Occidental',
    'ES': 'España',
    'ET': 'Etiopía',
    'FI': 'Finlandia',
    'FJ': 'Fiji',
    'FK': 'Islas Malvinas',
    'FM': 'Micronesia',
    'FO': 'Islas Faroe',
    'FR': 'Francia',
    'GA': 'Gabón',
    'GB': 'Gran Bretaña (Reino Unido)',
    'GD': 'Granada',
    'GF': 'Guyana Francesa',
    'GH': 'Ghana',
    'GI': 'Gibraltar',
    'GJ': 'Groenlandia',
    'GM': 'Gambia',
    'GN': 'Guinea',
    'GP': 'Guadalupe',
    'GQ': 'Guinea Ecuatorial',
    'GR': 'Grecia',
    'GT': 'Guatemala',
    'GU': 'Guam',
    'GW': 'Guinea Bissau',
    'GY': 'República de Guyana',
    'GZ': 'Islas de Guernesey, Jersey, Alderney, Isla Great Sark, Herm, Little Sark, Berchou, Jethou, Lihou (Islas del Canal)',
    'HK': 'Hong Kong',
    'HM': 'Islas Heard and Mc Donald',
    'HN': 'República de Honduras',
    'HT': 'Haití',
    'HU': 'Hungría',
    'ID': 'Indonesia',
    'IE': 'Irlanda',
    'IH': 'Isla del Hombre',
    'IL': 'Israel',
    'IN': 'India',
    'IO': 'Territorio Británico en el Océano Indico',
    'IP': 'Islas Pacífico',
    'IQ': 'Iraq',
    'IR': 'Irán',
    'IS': 'Islandia',
    'IT': 'Italia',
    'JM': 'Jamaica',
    'JO': 'Reino Hachemita de Jordania',
    'JP': 'Japón',
    'KE': 'Kenia',
    'KH': 'Campuchea Democrática',
    'KI': 'Kiribati',
    'KM': 'Comoros',
    'KN': 'San Kitts',
    'KP': 'República Democrática de Corea',
    'KR': 'República de Corea',
    'KW': 'Estado de Kuwait',
    'KY': 'Islas Caimán',
    'LA': 'República Democrática de Laos',
    'LB': 'Líbano',
    'LC': 'Santa Lucía',
    'LI': 'Principado de Liechtenstein',
    'LK': 'República Socialista Democrática de Sri Lanka',
    'LN': 'Labuán',
    'LR': 'República de Liberia',
    'LS': 'Lesotho',
    'LU': 'Gran Ducado de Luxemburgo',
    'LY': 'Libia',
    'MA': 'Marruecos',
    'MC': 'Principado de Mónaco',
    'MD': 'Madeira',
    'MG': 'Madagascar',
    'MH': 'República de las Islas Marshall',
    'ML': 'Malí',
    'MN': 'Mongolia',
    'MO': 'Macao',
    'MP': 'Islas Marianas del Noreste',
    'MQ': 'Martinica',
    'MR': 'Mauritania',
    'MS': 'Monserrat',
    'MT': 'Malta',
    'MU': 'República de Mauricio',
    'MV': 'República de Maldivas',
    'MW': 'Malawi',
    'MY': 'Malasia',
    'MZ': 'Mozambique',
    'NA': 'República de Namibia',
    'NC': 'Nueva Caledonia',
    'NE': 'Níger',
    'NF': 'Isla de Norfolk',
    'NG': 'Nigeria',
    'NI': 'Nicaragua',
    'NL': 'Holanda',
    'NO': 'Noruega',
    'NP': 'Nepal',
    'NR': 'República de Nauru',
    'NT': 'Zona Neutral',
    'NU': 'Niue',
    'NV': 'Nevis',
    'NZ': 'Nueva Zelandia',
    'OM': 'Sultanía de Omán',
    'PA': 'República de Panamá',
    'PE': 'Perú',
    'PF': 'Polinesia Francesa',
    'PG': 'Papúa Nueva Guinea',
    'PH': 'Filipinas',
    'PK': 'Pakistán',
    'PL': 'Polonia',
    'PM': 'Isla de San Pedro y Miguelón',
    'PN': 'Pitcairn',
    'PR': 'Estado Libre Asociado de Puerto Rico',
    'PT': 'Portugal',
    'PU': 'Patau',
    'PW': 'Palau',
    'PY': 'Paraguay',
    'QA': 'Estado de Quatar',
    'QB': 'Isla Qeshm',
    'RE': 'Reunión',
    'RO': 'Rumania',
    'RW': 'Rhuanda',
    'SA': 'Arabia Saudita',
    'SB': 'Islas Salomón',
    'SC': 'Seychelles Islas',
    'SD': 'Sudán',
    'SE': 'Suecia',
    'SG': 'Singapur',
    'SH': 'Santa Elena',
    'SI': 'Archipiélago de Svalbard',
    'SJ': 'Islas Svalbard and Jan Mayen',
    'SK': 'Sark',
    'SL': 'Sierra Leona',
    'SM': 'Serenísima República de San Marino',
    'SN': 'Senegal',
    'SO': 'Somalia',
    'SR': 'Surinam',
    'ST': 'Sao Tome and Príncipe',
    'SU': 'Países de la Ex-U.R.S.S., excepto Ucrania y Bielorusia',
    'SV': 'El Salvador',
    'SW': 'República de Seychelles',
    'SY': 'Siria',
    'SZ': 'Reino de Swazilandia',
    'TC': 'Islas Turcas y Caicos',
    'TD': 'Chad',
    'TF': 'Territorios Franceses del Sureste',
    'TG': 'Togo',
    'TH': 'Thailandia',
    'TK': 'Tokelau',
    'TN': 'República de Túnez',
    'TO': 'Reino de Tonga',
    'TP': 'Timor Este',
    'TR': 'Trieste',
    'TS': 'Tristán Da Cunha',
    'TT': 'República de Trinidad y Tobago',
    'TU': 'Turquía',
    'TV': 'Tuvalu',
    'TW': 'Taiwán',
    'TZ': 'Tanzania',
    'UA': 'Ucrania',
    'UG': 'Uganda',
    'UM': 'Islas menores alejadas de los Estados Unidos',
    'US': 'Estados Unidos de América',
    'UY': 'República Oriental del Uruguay',
    'VA': 'El Vaticano',
    'VC': 'San Vicente y Las Granadinas',
    'VE': 'Venezuela',
    'VG': 'Islas Vírgenes Británicas',
    'VI': 'Islas Vírgenes de Estados Unidos de América',
    'VN': 'Vietnam',
    'VU': 'República de Vanuatu',
    'WF': 'Islas Wallis y Funtuna',
    'XX': 'Otro',
    'YD': 'Yemen Democrática',
    'YE': 'República del Yemen',
    'YU': 'Países de la Ex-Yugoslavia',
    'ZA': 'Sudáfrica',
    'ZC': 'Zona Especial Canaria',
    'ZM': 'Zambia',
    'ZO': 'Zona Libre Ostrava',
    'ZR': 'Zaire',
    'ZW': 'Zimbawe',
}
PERIODIC = {
    '0': 'Sin Seleccion',
    '1': 'Mensual',
    '2': 'Bimestral',
    '3': 'Trimestral',
    '4': 'Cuatrimestral',
    '5': 'Semestral',
}
PERIODO = {
    "00": "Sin Seleccion",
    "01": "Enero",
    "02": "Febrero",
    "03": "Marzo",
    "04": "Abril",
    "05": "Mayo",
    "06": "Junio",
    "07": "Julio",
    "08": "Agosto",
    "09": "Septiembre",
    "10": "Octubre",
    "11": "Noviembre",
    "12": "Diciembre",
    "13": "Enero - Febrero",
    "14": "Marzo - Abril",
    "15": "Mayo - Junio",
    "16": "Julio - Agosto",
    "17": "Septiembre - Octubre",
    "18": "Noviembre - Diciembre",
    "19": "Enero - Marzo",
    "20": "Abril - Junio",
    "21": "Julio - Septiembre",
    "22": "Octubre - Diciembre",
    "23": "Enero - Abril",
    "24": "Mayo - Agosto",
    "25": "Septiembre - Diciembre",
    "26": "Enero - Junio",
    "27": "Julio - Diciembre",
}
PRESENTA = {
    '0': 'La presenta sin operaciones',
    '1': 'La presenta con datos',
}
TIP_DECLA = {
    '1': 'Normal',
    '2': 'Complementaria',
}
TIP_OPERA = {
    '03': 'Prestación de Servicios Profesionales',
    '06': 'Arrendamiento de Inmuebles',
    '85': 'Otros',
}
TIPO_TERC = {
    '04': 'Proveedor Nacional',
    '05': 'Proveedor Extranjero',
    '15': 'Proveedor Global',
}



from ...utils import StrEnum


class FormaPago(StrEnum):
    EFECTIVO = '01'
    CHEQUE_NOMINATIVO = '02'
    TRANSFERENCIA_ELECTRONICA_DE_FONDOS = '03'
    TARJETA_DE_CREDITO = '04'
    MONEDERO_ELECTRONICO = '05'
    DINERO_ELECTRONICO = '06'
    VALES_DE_DESPENSA = '08'
    DACION_EN_PAGO = '12'
    PAGO_POR_SUBROGACION = '13'
    PAGO_POR_CONSIGNACION = '14'
    CONDONACION = '15'
    COMPENSACION = '17'
    NOVACION = '23'
    CONFUSION = '24'
    REMISION_DE_DEUDA = '25'
    PRESCRIPCION_O_CADUCIDAD = '26'
    A_SATISFACCION_DEL_ACREEDOR = '27'
    TARJETA_DE_DEBITO = '28'
    TARJETA_DE_SERVICIOS = '29'
    APLICACION_DE_ANTICIPOS = '30'
    INTERMEDIARIO_PAGOS = '31'
    POR_DEFINIR = '99'
    TAREJTAS_DIGITALES = '07'
    BIENES = '09'
    SERVICIO = '10'
    POR_CUENTA_DE_TERCERO = '11'
    CANCELACION = '16'
    NA = '98'


class Impuesto(StrEnum):
    ISR = '001'
    IVA = '002'
    IEPS = '003'


class MetodoPago(StrEnum):
    PAGO_EN_UNA_SOLA_EXHIBICION = 'PUE'
    PAGO_EN_PARCIALIDADES_O_DIFERIDO = 'PPD'


class RegimenFiscal(StrEnum):
    GENERAL_DE_LEY_PERSONAS_MORALES = '601'
    PERSONAS_MORALES_CON_FINES_NO_LUCRATIVOS = '603'
    SUELDOS_Y_SALARIOS_E_INGRESOS_ASIMILADOS_A_SALARIOS = '605'
    ARRENDAMIENTO = '606'
    REGIMEN_DE_ENAJENACION_O_ADQUISICION_DE_BIENES = '607'
    DEMAS_INGRESOS = '608'
    CONSOLIDACION = '609'
    RESIDENTES_EN_EL_EXTRANJERO_SIN_ESTABLECIMIENTO_PERMANENTE_EN_MEXICO = '610'
    INGRESOS_POR_DIVIDENDOS_SOCIOS_Y_ACCIONISTAS = '611'
    PERSONAS_FISICAS_CON_ACTIVIDADES_EMPRESARIALES_Y_PROFESIONALES = '612'
    INGRESOS_POR_INTERESES = '614'
    REGIMEN_DE_LOS_INGRESOS_POR_OBTENCION_DE_PREMIOS = '615'
    SIN_OBLIGACIONES_FISCALES = '616'
    SOCIEDADES_COOPERATIVAS_DE_PRODUCCION_QUE_OPTAN_POR_DIFERIR_SUS_INGRESOS = '620'
    INCORPORACION_FISCAL = '621'
    ACTIVIDADES_AGRICOLAS_GANADERAS_SILVICOLAS_Y_PESQUERAS = '622'
    OPCIONAL_PARA_GRUPOS_DE_SOCIEDADES = '623'
    COORDINADOS = '624'
    REGIMEN_DE_LAS_ACTIVIDADES_EMPRESARIALES_CON_INGRESOS_A_TRAVES_DE_PLATAFORMAS_TECNOLOGICAS = '625'
    REGIMEN_SIMPLIFICADO_DE_CONFIANZA = '626'
    HIDROCARBUROS = '628'
    DE_LOS_REGIMENES_FISCALES_PREFERENTES_Y_DE_LAS_EMPRESAS_MULTINACIONALES = '629'
    ENAJENACION_DE_ACCIONES_EN_BOLSA_DE_VALORES = '630'


class TipoDeComprobante(StrEnum):
    INGRESO = 'I'
    EGRESO = 'E'
    TRASLADO = 'T'
    NOMINA = 'N'
    PAGO = 'P'


class TipoFactor(StrEnum):
    TASA = 'Tasa'
    CUOTA = 'Cuota'
    EXENTO = 'Exento'


class TipoRelacion(StrEnum):
    NOTA_DE_CREDITO_DE_LOS_DOCUMENTOS_RELACIONADOS = '01'
    NOTA_DE_DEBITO_DE_LOS_DOCUMENTOS_RELACIONADOS = '02'
    DEVOLUCION_DE_MERCANCIA_SOBRE_FACTURAS_O_TRASLADOS_PREVIOS = '03'
    SUSTITUCION_DE_LOS_CFDI_PREVIOS = '04'
    TRASLADOS_DE_MERCANCIAS_FACTURADOS_PREVIAMENTE = '05'
    FACTURA_GENERADA_POR_LOS_TRASLADOS_PREVIOS = '06'
    CFDI_POR_APLICACION_DE_ANTICIPO = '07'
    FACTURA_GENERADA_POR_PAGOS_EN_PARCIALIDADES = '08'
    FACTURA_GENERADA_POR_PAGOS_DIFERIDOS = '09'


class UsoCFDI(StrEnum):
    ADQUISICION_DE_MERCANCIAS = 'G01'
    DEVOLUCIONES_DESCUENTOS_O_BONIFICACIONES = 'G02'
    GASTOS_EN_GENERAL = 'G03'
    CONSTRUCCIONES = 'I01'
    MOBILIARIO_Y_EQUIPO_DE_OFICINA_POR_INVERSIONES = 'I02'
    EQUIPO_DE_TRANSPORTE = 'I03'
    EQUIPO_DE_COMPUTO_Y_ACCESORIOS = 'I04'
    DADOS_TROQUELES_MOLDES_MATRICES_Y_HERRAMENTAL = 'I05'
    COMUNICACIONES_TELEFONICAS = 'I06'
    COMUNICACIONES_SATELITALES = 'I07'
    OTRA_MAQUINARIA_Y_EQUIPO = 'I08'
    HONORARIOS_MEDICOS_DENTALES_Y_GASTOS_HOSPITALARIOS = 'D01'
    GASTOS_MEDICOS_POR_INCAPACIDAD_O_DISCAPACIDAD = 'D02'
    GASTOS_FUNERALES = 'D03'
    DONATIVOS = 'D04'
    INTERESES_REALES_EFECTIVAMENTE_PAGADOS_POR_CREDITOS_HIPOTECARIOS_CASA_HABITACION = 'D05'
    APORTACIONES_VOLUNTARIAS_AL_SAR = 'D06'
    PRIMAS_POR_SEGUROS_DE_GASTOS_MEDICOS = 'D07'
    GASTOS_DE_TRANSPORTACION_ESCOLAR_OBLIGATORIA = 'D08'
    DEPOSITOS_EN_CUENTAS_PARA_EL_AHORRO_PRIMAS_QUE_TENGAN_COMO_BASE_PLANES_DE_PENSIONES = 'D09'
    PAGOS_POR_SERVICIOS_EDUCATIVOS_COLEGIATURAS = 'D10'
    SIN_EFECTOS_FISCALES = 'S01'
    PAGOS = 'CP01'
    NOMINA = 'CN01'
    POR_DEFINIR = 'P01'


class Exportacion(StrEnum):
    NO_APLICA = '01'
    DEFINITIVA_CON_CLAVE_A1 = '02'
    TEMPORAL = '03'
    DEFINITIVA_CON_CLAVE_DISTINTA_A_A1_O_CUANDO_NO_EXISTE_ENAJENACION_EN_TERMINOS_DEL_CFF = '04'


class ObjetoImp(StrEnum):
    NO_OBJETO_DE_IMPUESTO = '01'
    SI_OBJETO_DE_IMPUESTO = '02'
    SI_OBJETO_DEL_IMPUESTO_Y_NO_OBLIGADO_AL_DESGLOSE = '03'
    SI_OBJETO_DEL_IMPUESTO_Y_NO_CAUSA_IMPUESTO = '04'





import os
from datetime import date, datetime
from decimal import Decimal
from io import BytesIO
from typing import Sequence
from zipfile import ZipFile, ZipInfo

from .catalog import *
from .code import *
from .utils import _format_rfc, DIOTWriter, encrypt_triple_des, period_code, catalog_code
from .. import RFC, RFCType, CURP, iterate, Certificate
from ..ans1e import *
from ..printer import Representable

__all__ = [
    'DIOT',
    'DatosIdentificacion',
    'ProveedorTercero',
    'DatosComplementaria',
    'TipoOperacion',
    'TipoTercero',
    'Periodo',
    'Pais'
]

from ..zip import zip_create, ZipData

current_dir = os.path.dirname(__file__)

VERSION_DEMC = 10
CODES = "0123456789ABCDEFGHIJKLMNOPQRSTUV"

CERTIFICATE = Certificate.load_certificate(
    certificate=open(os.path.join(current_dir, 'De_srv_x509.CER'), 'rb').read()
)


class DatosIdentificacion:
    def __init__(
            self,
            rfc: str | RFC,
            ejercicio: int,
            curp: str = None,
            nombre: str = None,
            apellido_paterno: str = None,
            apellido_materno: str = None,
            razon_social: str = None):

        self.rfc = RFC(rfc)
        self.ejercicio = ejercicio
        self.curp = CURP(curp) if curp else None
        self.nombre = nombre.upper() if nombre else None
        self.apellido_paterno = apellido_paterno.upper() if apellido_paterno else None
        self.apellido_materno = apellido_materno.upper() if apellido_materno else None
        self.razon_social = razon_social.upper() if razon_social else None

    def to_dict(self):
        return {
            "RFC": self.rfc,
            "CURP": self.curp,
            "Ejercicio": self.ejercicio,
            "ApellidoPaterno": self.apellido_paterno,
            "ApellidoMaterno": self.apellido_materno,
            "Nombre": self.nombre,
            "RazonSocial": self.razon_social
        }

    def render(self, w: DIOTWriter):
        w(10001, 0, 0, _format_rfc(self.rfc))
        if self.rfc.type == RFCType.FISICA:
            w(10002, 0, 0, self.curp)

        w(10021, 0, 0, self.ejercicio)
        if self.rfc.type == RFCType.FISICA:
            w(10003, 0, 0, self.apellido_paterno)
            w(10004, 0, 0, self.apellido_materno)
            w(10005, 0, 0, self.nombre)
        else:
            w(10006, 0, 0, self.razon_social)


class ProveedorTercero:
    def __init__(
            self,
            tipo_tercero: TipoTercero,
            tipo_operacion: TipoOperacion,
            rfc: str | RFC = None,
            id_fiscal: str = None,
            nombre_extranjero: str = None,
            pais: Pais = None,
            nacionalidad: str = None,
            iva16: int = None,
            iva16_na: int = None,
            iva_rfn: int = None,
            iva_rfn_na: int = None,
            iva_import16: int = None,
            iva_import16_na: int = None,
            iva_import_exento: int = None,
            iva0: int = None,
            iva_exento: int = None,
            retenido: int = None,
            devoluciones: int = None):
        """
        :param tipo_tercero:
        :param tipo_operacion:
        :param rfc:
        :param id_fiscal: Numero de ID Fiscal
        :param nombre_extranjero: Nombre del Extranjero
        :param pais: Pais de residencia
        :param nacionalidad: Nacionalidad
        :param iva16: Total de los actos o actividades pagados a la tasa 16% de IVA
        :param iva16_na: Total del IVA pagado no acreditable a la tasa 16%
        :param iva_rfn: Total de los actos o actividades pagados sujeto al estimulo de la region fronteriza norte
        :param iva_rfn_na: Total del IVA pagado no acreditable sujeto al estimulo de la region fronteriza norte
        :param iva_import16: Total de los actos o actividades pagados en la importacion de bienes y servicios a la tasa 16%
        :param iva_import16_na: Total del IVA pagado no acreditable en la importacion de bienes y servicios a la tasa 16%
        :param iva_import_exento: Total de los actos o actividades pagados en la importacion de bienes y servicios por los que no se pagara IVA (Exentos)
        :param iva0: Total de los actos o actividades pagados a la tasa 0% de IVA
        :param iva_exento: Total de los actos o actividades pagados por los que no se pagara IVA (Exentos)
        :param retenido: Total del IVA Retenido por el contribuyente
        :param devoluciones: Total del IVA correspondiente a las devoluciones, descuentos y bonificaciones sobre compras
        """

        self.tipo_tercero = tipo_tercero
        self.tipo_operacion = tipo_operacion
        self.rfc = RFC(rfc) if rfc else None
        self.id_fiscal = id_fiscal
        self.nombre_extranjero = nombre_extranjero
        self.pais = pais
        self.nacionalidad = nacionalidad

        self.iva16 = iva16
        self.iva16_na = iva16_na
        self.iva_rfn = iva_rfn
        self.iva_rfn_na = iva_rfn_na
        self.iva_import16 = iva_import16
        self.iva_import16_na = iva_import16_na
        self.iva_import_exento = iva_import_exento
        self.iva0 = iva0
        self.iva_exento = iva_exento
        self.retenido = retenido
        self.devoluciones = devoluciones

        match self.tipo_tercero:
            case TipoTercero.PROVEEDOR_NACIONAL:
                if not self.rfc:
                    raise ValueError("RFC is required for NACIONAL")
            case TipoTercero.PROVEEDOR_EXTRANJERO:
                pass
            case TipoTercero.PROVEEDOR_GLOBAL:
                if self.rfc:
                    raise ValueError("RFC must be empty for GLOBAL")

    def to_dict(self):
        return {
            "TipoTercero": catalog_code(TIPO_TERC, self.tipo_tercero),
            "TipoOperacion": catalog_code(TIP_OPERA, self.tipo_operacion),
            "RFC": self.rfc,
            "NumeroIDFiscal": self.id_fiscal,
            "NombreExtranjero": self.nombre_extranjero,
            "PaisResidencia": catalog_code(PAISES, self.pais),
            "Nacionalidad": self.nacionalidad,

            "ActividadesIVATasa16": self.iva16,
            "NoAcreditableIVATasa16": self.iva16_na,
            "ActividadesRegionFronterizaNorte": self.iva_rfn,
            "NoAcreditableRegionFronterizaNorte": self.iva_rfn_na,
            "ActividadesImportacionIVATasa16": self.iva_import16,
            "NoAcreditableImportacionIVATasa16": self.iva_import16_na,
            "ActividadesImportacionExento": self.iva_import_exento,
            "ActividadesIVATasa0": self.iva0,
            "ActividadesIVAExento": self.iva_exento,
            "Retenido": self.retenido,
            "Devoluciones": self.devoluciones
        }

    def render(self, w: DIOTWriter):
        w(200261, 1, 1, int(self.tipo_tercero))
        w(200361, 1, 1, int(self.tipo_operacion))
        if self.id_fiscal:
            w(200561, 1, 1, self.id_fiscal)
        if self.nombre_extranjero:
            w(200661, 1, 1, self.nombre_extranjero)
        if self.pais:
            w(200761, 1, 1, self.pais)
        if self.nacionalidad:
            w(200861, 1, 1, self.nacionalidad)
        if self.rfc:
            w(200461, 1, 1, _format_rfc(self.rfc))

        if self.iva16 is not None:
            w(301061, 1, 1, self.iva16)
        if self.iva16_na is not None:
            w(301161, 1, 1, self.iva16_na)
        if self.iva_rfn is not None:
            w(302461, 1, 1, self.iva_rfn)
        if self.iva_rfn_na is not None:
            w(301361, 1, 1, self.iva_rfn_na)
        if self.iva_import16 is not None:
            w(301461, 1, 1, self.iva_import16)
        if self.iva_import16_na is not None:
            w(301561, 1, 1, self.iva_import16_na)
        if self.iva_import_exento is not None:
            w(301861, 1, 1, self.iva_import_exento)
        if self.iva0 is not None:
            w(301961, 1, 1, self.iva0)
        if self.iva_exento is not None:
            w(302061, 1, 1, self.iva_exento)
        if self.retenido is not None:
            w(302161, 1, 1, self.retenido)
        if self.devoluciones is not None:
            w(302261, 1, 1, self.devoluciones)


class DatosComplementaria:
    def __init__(
            self,
            folio_anterior: str,
            fecha_presentacion_anterior: date
    ):
        self.folio_anterior = folio_anterior
        self.fecha_presentacion_anterior = fecha_presentacion_anterior

    def render(self, w: DIOTWriter):
        w(361, 0, 0, self.folio_anterior)
        w(461, 0, 0, self.fecha_presentacion_anterior.strftime('%Y%m%d'))

    def to_dict(self):
        return {
            "FolioAnterior": self.folio_anterior,
            "FechaPresentacionAnterior": self.fecha_presentacion_anterior
        }


class Totales:
    def __init__(self, providers: Sequence[ProveedorTercero]):
        self.operations_count = len(providers)

        self.iva16 = sum(p.iva16 or 0 for p in providers)
        self.iva16_na = sum(p.iva16_na or 0 for p in providers)
        self.iva_rfn = sum(p.iva_rfn or 0 for p in providers)
        self.iva_rfn_na = sum(p.iva_rfn_na or 0 for p in providers)
        self.iva_import16 = sum(p.iva_import16 or 0 for p in providers)
        self.iva_import16_na = sum(p.iva_import16_na or 0 for p in providers)
        self.iva_import_exento = sum(p.iva_import_exento or 0 for p in providers)
        self.iva0 = sum(p.iva0 or 0 for p in providers)
        self.iva_exento = sum(p.iva_exento or 0 for p in providers)
        self.retenido = sum(p.retenido or 0 for p in providers)
        self.devoluciones = sum(p.devoluciones or 0 for p in providers)

        # N102561=®((((N101261*16)/100)+((N101461*11)/100)+((N102761*15)/100)+((N102961*16)/100)*0.5)-(10/1000))
        # Note: N102961 es N102861
        self.total = round(self.iva16 * Decimal(".16") + self.iva_rfn * Decimal(".08") - Decimal(".01"))

        # N102661=®(((N101661*16)/100)+((N101861*11)/100)-(10/1000))
        self.total_import = round(self.iva_import16 * Decimal(".16") - Decimal(".01"))

    def to_dict(self):
        return {
            "TotalOperaciones": self.operations_count,

            "ActividadesIVATasa16": self.iva16,
            "NoAcreditableIVATasa16": self.iva16_na,
            "ActividadesRegionFronterizaNorte": self.iva_rfn,
            "NoAcreditableRegionFronterizaNorte": self.iva_rfn_na,
            "ActividadesImportacionIVATasa16": self.iva_import16,
            "NoAcreditableImportacionIVATasa16": self.iva_import16_na,
            "ActividadesImportacionExento": self.iva_import_exento,
            "ActividadesIVATasa0": self.iva0,
            "ActividadesIVAExento": self.iva_exento,
            "Retenido": self.retenido,
            "Devoluciones": self.devoluciones,

            "TotalExceptoImportaciones": self.total,
            "TotalImportaciones": self.total_import
        }

    def render(self, w: DIOTWriter):
        w(100161, 1, 0, self.operations_count)
        w(101261, 1, 0, self.iva16)
        w(101361, 1, 0, self.iva16_na)
        w(101461, 1, 0, 0)
        w(102861, 1, 0, self.iva_rfn)
        w(101561, 1, 0, self.iva_rfn_na)
        w(101661, 1, 0, self.iva_import16)
        w(101761, 1, 0, self.iva_import16_na)
        w(101861, 1, 0, 0)
        w(101961, 1, 0, 0)
        w(102061, 1, 0, self.iva_import_exento)
        w(102161, 1, 0, self.iva0)
        w(102261, 1, 0, self.iva_exento)
        w(102361, 1, 0, self.retenido)
        w(102461, 1, 0, self.devoluciones)
        w(102561, 1, 0, self.total)
        w(102661, 1, 0, self.total_import)


class DatosGenerales:
    def __init__(
            self,
            declaracion: str,
            periodo: Periodo,
            complementaria: DatosComplementaria = None,
    ):
        self.periodo = periodo
        self.periodicidad = period_code(self.periodo)

        self.complementaria = complementaria
        if complementaria:
            self.type = "2"
        else:
            self.type = "1"

        self.declaracion = declaracion

    def render(self, w: DIOTWriter):
        w(161, 0, 0, self.declaracion)
        w(261, 0, 0, self.type)
        if self.complementaria:
            self.complementaria.render(w)
        w(661, 0, 0, int(self.periodicidad))
        w(561, 0, 0, int(self.periodo))

    def to_dict(self):
        return {
            "Declaracion": catalog_code(PRESENTA, self.declaracion),
            "TipoDeclaracion": catalog_code(TIP_DECLA, self.type),
            "Complementaria": self.complementaria.to_dict() if self.complementaria else None,
            "Periodicidad": catalog_code(PERIODIC, self.periodicidad),
            "Periodo": catalog_code(PERIODO, self.periodo)
        }


class DIOT(Representable):
    tag = "DIOT"

    def __init__(
            self,
            datos_identificacion: DatosIdentificacion,
            periodo: Periodo,
            complementaria: DatosComplementaria = None,
            proveedores: ProveedorTercero | Sequence[ProveedorTercero] = None
    ):
        proveedores = iterate(proveedores)

        self.datos_identificacion = datos_identificacion
        self.datos_generales = DatosGenerales(
            declaracion="1" if proveedores else "0",
            periodo=periodo,
            complementaria=complementaria
        )
        self.proveedores = proveedores
        self.totales = Totales(proveedores)

    def to_dict(self):
        return {
            "DatosIdentificacion": self.datos_identificacion.to_dict(),
            "DatosGenerales": self.datos_generales.to_dict(),
            "Proveedores": [p.to_dict() for p in self.proveedores],
            "Totales": self.totales.to_dict(),
        }

    def render(self, w: DIOTWriter):
        w(40006, 0, 0, '20062')
        w(10040, 0, 0, '081')

        self.datos_identificacion.render(w)
        self.datos_generales.render(w)

        if self.proveedores:
            self.totales.render(w)
            for i, p in enumerate(self.proveedores, start=1):
                w(200161, 1, 1, i)
                p.render(w)

        w.end()

    def _tmp_filename(self):
        rfc = self.datos_identificacion.rfc

        if len(rfc) == 12:
            rfc = f"_{rfc}"

        v_dem = "0"
        n_formulario = 'DOT'
        ver_clte = CODES[VERSION_DEMC]
        ver_form = CODES[10]
        mes_inicial = CODES[1]
        mes_final = CODES[12]
        period = CODES[self.datos_identificacion.ejercicio % 100]

        x_date = datetime.now()
        anio = CODES[x_date.year % 100]
        mes = CODES[x_date.month]
        dia = CODES[x_date.day]
        hora = x_date.strftime("%H%M")

        suffix = "2" if len(self.proveedores) > 40000 else "1"

        return f"{rfc}{v_dem}{n_formulario}{ver_clte}{ver_form}{period}{mes_inicial}{period}{mes_final}{anio}{mes}{dia}{hora}{suffix}"

    def plain_bytes(self) -> bytes:
        with BytesIO() as b:
            self.plain_write(b)
            return b.getvalue()

    def plain_write(self, target):
        w = DIOTWriter(target)
        self.render(w)

    def _zip_bytes(self, tmp_filename) -> bytes:
        with BytesIO() as b:
            self._zip_write(b, tmp_filename)
            return b.getvalue()

    def _zip_write(self, target: BytesIO, tmp_filename):
        zip_create(target, [
            ZipData(f"C:\\{tmp_filename}.txt", lambda s: self.render(DIOTWriter(s)))
        ])

    def _encrypted_bytes(self, tmp_filename) -> bytes:
        with BytesIO() as b:
            self._encrypted_write(b, tmp_filename)
            return b.getvalue()

    def _encrypted_write(self, target, tmp_filename):
        with BytesIO() as zip_data:
            self._zip_write(zip_data, tmp_filename)
            algo, key, iv, data_encrypted = encrypt_triple_des(zip_data)
            issuer = CERTIFICATE.certificate.get_issuer().der()
            enc_key = CERTIFICATE.encrypt(key)

        # SERIALIZE ANS1 FILE
        e = Ans1Encoder()
        with e.seq():
            e.oid("1.2.840.113549.1.7.3")
            with e.enter(nr=0, cls=Classes.Context):
                with e.seq():
                    e(0)
                    with e.set():
                        with e.seq():
                            e(0)
                            with e.seq():
                                e.write(issuer)
                                e(CERTIFICATE.serial_number)
                            with e.seq():
                                e.oid("1.2.840.113549.1.1.1")
                                e()
                            e(enc_key)
                    with e.seq():
                        e.oid("1.2.840.113549.1.7.1")
                        with e.seq():
                            e.oid(algo)
                            e(iv)
                        e(data_encrypted.getvalue(), nr=0, cls=Classes.Context)
        e.stream(target)

    def generate_package(self, dirname: str = None) -> str:
        """
        Generate Package, return filename
        """
        tmp_filename = self._tmp_filename()
        if dirname:
            filename = os.path.join(dirname, tmp_filename + ".dec")
        else:
            filename = tmp_filename + ".dec"
        with open(filename, 'wb') as f:
            self._encrypted_write(f, tmp_filename)
        return filename



"""notariospublicos http://www.sat.gob.mx/notariospublicos"""
from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ... import CFDI, XElement, ScalarMap


class DatosAdquirienteCopSC(ScalarMap):
    """
    Nodo para capturar los datos de un adquiriente o de un propietario o poseedor en caso de Copropiedad o Sociedad Conyugal
    
    :param nombre: Atributo requerido para expresar el nombre, denominación o razón social de cada adquiriente o de cada propietario o poseedor en caso de servidumbres de paso.
    :param rfc: Atributo requerido para la Clave del Registro Federal de Contribuyentes sin guiones o espacios, correspondiente a cada adquiriente o de cada propietario o poseedor en caso de servidumbres de paso.
    :param porcentaje: Porcentaje que le corresponde en la copropiedad a cada adquiriente o de cada propietario o poseedor en caso de servidumbres de paso.
    :param apellido_paterno: Atributo opcional para expresar el apellido paterno de cada adquiriente o de cada propietario o poseedor en caso de servidumbres de paso.
    :param apellido_materno: Atributo opcional para expresar el apellido materno de cada adquirientes o de cada propietario o poseedor en caso de servidumbres de paso.
    :param curp: Atributo opcional para expresar la CURP de cada adquiriente o de cada propietario o poseedor en caso de servidumbres de paso.
    """
    
    def __init__(
            self,
            nombre: str,
            rfc: str,
            porcentaje: Decimal | int,
            apellido_paterno: str = None,
            apellido_materno: str = None,
            curp: str = None,
    ): 
        super().__init__({
            'Nombre': nombre,
            'RFC': rfc,
            'Porcentaje': porcentaje,
            'ApellidoPaterno': apellido_paterno,
            'ApellidoMaterno': apellido_materno,
            'CURP': curp,
        })
        

class DatosUnAdquiriente(ScalarMap):
    """
    Nodo para capturar los datos del adquiriente o del propietario o poseedor en caso de ser solo uno.
    
    :param nombre: Atributo requerido para expresar el nombre, denominación o razón social del adquiriente o del propietario o poseedor del bien dominante o del pagador de la indemnización o contraprestación en el caso de servidumbres de paso.
    :param rfc: Atributo requerido para la Clave del Registro Federal de Contribuyentes sin guiones o espacios, del adquiriente o del propietario o poseedor del bien dominante o del pagador de la indemnización o contraprestación en el caso de servidumbres de paso.
    :param apellido_paterno: Atributo opcional para expresar el apellido paterno del adquiriente o del propietario o poseedor del bien dominante o del pagador de la indemnización o contraprestación en el caso de servidumbres de paso.
    :param apellido_materno: Atributo opcional para expresar el apellido materno del adquiriente o del propietario o poseedor del bien dominante o del pagador de la indemnización o contraprestación en el caso de servidumbres de paso.
    :param curp: Atributo opcional para expresar la CURP del adquiriente o del propietario o poseedor del bien dominante o del pagador de la indemnización o contraprestación en el caso de servidumbres de paso.
    """
    
    def __init__(
            self,
            nombre: str,
            rfc: str,
            apellido_paterno: str = None,
            apellido_materno: str = None,
            curp: str = None,
    ): 
        super().__init__({
            'Nombre': nombre,
            'RFC': rfc,
            'ApellidoPaterno': apellido_paterno,
            'ApellidoMaterno': apellido_materno,
            'CURP': curp,
        })
        

class DatosAdquiriente(ScalarMap):
    """
    Nodo para capturar los datos del adquiriente, adquirientes o propietario, o propietarios o poseedores, en caso de servidumbres de paso.
    
    :param copro_soc_conyugal_e: Atributo requerido que expresa si es copropiedad o sociedad conyugal
    :param datos_un_adquiriente: Nodo para capturar los datos del adquiriente o del propietario o poseedor en caso de ser solo uno.
    :param datos_adquirientes_cop_sc: Nodo para capturar los datos de los adquirientes o propietarios o poseedores del bien dominante en caso de Copropiedad o Sociedad Conyugal
    """
    
    def __init__(
            self,
            copro_soc_conyugal_e: str,
            datos_un_adquiriente: DatosUnAdquiriente | dict = None,
            datos_adquirientes_cop_sc: DatosAdquirienteCopSC | dict | Sequence[DatosAdquirienteCopSC | dict] = None,
    ): 
        super().__init__({
            'CoproSocConyugalE': copro_soc_conyugal_e,
            'DatosUnAdquiriente': datos_un_adquiriente,
            'DatosAdquirientesCopSC': datos_adquirientes_cop_sc,
        })
        

class DatosEnajenanteCopSC(ScalarMap):
    """
    Nodo para capturar los datos de un enajenante o de los propietarios o poseedores tratándose de servidumbres de paso, en caso de Copropiedad o Sociedad Conyugal
    
    :param nombre: Atributo requerido para expresar el nombre (s) de cada enajenante o del propietario o poseedor del predio sirviente, en caso de servidumbres de paso.
    :param rfc: Atributo requerido para la Clave del Registro Federal de Contribuyentes sin guiones o espacios, correspondiente a cada enajenante o del propietario o poseedor del predio sirviente, en caso de servidumbres de paso.
    :param porcentaje: Porcentaje que le corresponde en la copropiedad a cada enajenante o del propietario o poseedor del predio sirviente, en caso de servidumbres de paso.
    :param apellido_paterno: Atributo opcional para expresar el apellido paterno de cada enajenante o del propietario o poseedor del predio sirviente, en caso de servidumbres de paso.
    :param apellido_materno: Atributo opcional para expresar el apellido materno de cada enajenante o del propietario o poseedor del predio sirviente, en caso de servidumbres de paso.
    :param curp: Atributo opcional para expresar la CURP de cada enajenante o del propietario o poseedor del predio sirviente, en caso de servidumbres de paso.
    """
    
    def __init__(
            self,
            nombre: str,
            rfc: str,
            porcentaje: Decimal | int,
            apellido_paterno: str = None,
            apellido_materno: str = None,
            curp: str = None,
    ): 
        super().__init__({
            'Nombre': nombre,
            'RFC': rfc,
            'Porcentaje': porcentaje,
            'ApellidoPaterno': apellido_paterno,
            'ApellidoMaterno': apellido_materno,
            'CURP': curp,
        })
        

class DatosUnEnajenante(ScalarMap):
    """
    Nodo para capturar los datos del enajenante o del propietario o poseedor del predio sirviente en caso de ser solo uno.
    
    :param nombre: Atributo requerido para expresar el nombre del enajenante o del propietario o poseedor del predio sirviente, en caso de servidumbres de paso.
    :param apellido_paterno: Atributo requerido para expresar el apellido paterno del enajenante o del propietario o poseedor del predio sirviente, en caso de servidumbres de paso.
    :param rfc: Atributo requerido para la Clave del Registro Federal de Contribuyentes sin guiones o espacios, del enajenante o del propietario o poseedor del predio sirviente, en caso de servidumbres de paso.
    :param curp: Atributo requerido para expresar la CURP del enajenante o del propietario o poseedor del predio sirviente, en caso de servidumbres de paso.
    :param apellido_materno: Atributo opcional para expresar el apellido materno del enajenante o del propietario o poseedor del predio sirviente, en caso de servidumbres de paso.
    """
    
    def __init__(
            self,
            nombre: str,
            apellido_paterno: str,
            rfc: str,
            curp: str,
            apellido_materno: str = None,
    ): 
        super().__init__({
            'Nombre': nombre,
            'ApellidoPaterno': apellido_paterno,
            'RFC': rfc,
            'CURP': curp,
            'ApellidoMaterno': apellido_materno,
        })
        

class DatosEnajenante(ScalarMap):
    """
    Nodo para capturar los datos del enajenante o enajenantes, o en el caso de servidumbres de paso del propietario o poseedores o propietarios o poseedores del predio sirviente.
    
    :param copro_soc_conyugal_e: Atributo requerido que expresa si es copropiedad o sociedad conyugal
    :param datos_un_enajenante: Nodo para capturar los datos del enajenante o del propietario o poseedor del predio sirviente en caso de ser solo uno.
    :param datos_enajenantes_cop_sc: Nodo para capturar los datos de los enajenantes o de los propietarios o poseedores tratándose de servidumbres de paso, en caso de Copropiedad o Sociedad Conyugal
    """
    
    def __init__(
            self,
            copro_soc_conyugal_e: str,
            datos_un_enajenante: DatosUnEnajenante | dict = None,
            datos_enajenantes_cop_sc: DatosEnajenanteCopSC | dict | Sequence[DatosEnajenanteCopSC | dict] = None,
    ): 
        super().__init__({
            'CoproSocConyugalE': copro_soc_conyugal_e,
            'DatosUnEnajenante': datos_un_enajenante,
            'DatosEnajenantesCopSC': datos_enajenantes_cop_sc,
        })
        

class DatosNotario(ScalarMap):
    """
    
    :param curp: Atributo requerido para expresar la CURP del notario
    :param num_notaria: Atributo requerido para indicar el número de la Notaria que realizar la operación.
    :param entidad_federativa: Entidad Federativa donde se ubica la Notaria conforme al catálogo publicado en el portal del SAT en internet.
    :param adscripcion: Atributo opcional que expresa el señalamiento del notario a la plaza a la que se encuentra adscrito
    """
    
    def __init__(
            self,
            curp: str,
            num_notaria: int,
            entidad_federativa: str,
            adscripcion: str = None,
    ): 
        super().__init__({
            'CURP': curp,
            'NumNotaria': num_notaria,
            'EntidadFederativa': entidad_federativa,
            'Adscripcion': adscripcion,
        })
        

class DatosOperacion(ScalarMap):
    """
    Nodo para definir los detalles de la operación.
    
    :param num_instrumento_notarial: Atributo requerido que indica el número del instrumento Notarial donde consta la operación
    :param fecha_inst_notarial: Atributo requerido que indica la fecha de firma del instrumento Notarial
    :param monto_operacion: Atributo requerido para expresar el monto de la contraprestación, indemnización o valor de la operación.
    :param subtotal: Atributo requerido para expresar el subtotal de la contraprestación, indemnización o su valor en la operación
    :param iva: Atributo requerido para expresar el IVA de la contraprestación, indemnización o su valor en la operación
    """
    
    def __init__(
            self,
            num_instrumento_notarial: int,
            fecha_inst_notarial: date,
            monto_operacion: Decimal | int,
            subtotal: Decimal | int,
            iva: Decimal | int,
    ): 
        super().__init__({
            'NumInstrumentoNotarial': num_instrumento_notarial,
            'FechaInstNotarial': fecha_inst_notarial,
            'MontoOperacion': monto_operacion,
            'Subtotal': subtotal,
            'IVA': iva,
        })
        

class DescInmueble(ScalarMap):
    """
    Nodo para describir el inmueble o inmuebles objeto del acto otorgado.
    
    :param tipo_inmueble: Atributo requerido para la expresión del tipo de inmueble enajenado o sujeto a servidumbre (sirviente) conforme al catálogo publicado en el portal del SAT en internet.
    :param calle: Este atributo requerido sirve para precisar la avenida, calle, camino o carretera donde se ubica el inmueble
    :param municipio: Atributo requerido que sirve para precisar el municipio o delegación (en el caso del Distrito Federal) en donde se da la ubicación del inmueble
    :param estado: Entidad Federativa donde se ubica el inmueble conforme al catálogo publicado en el portal del SAT en internet.
    :param pais: Atributo requerido que sirve para precisar el país donde se da la ubicación, conforme al catálogo publicado en el portal del SAT en internet. En caso de servidumbres de paso, siempre será México.
    :param codigo_postal: Atributo requerido que sirve para asentar el código postal en donde se da la ubicación del inmueble.
    :param no_exterior: Este atributo opcional sirve para expresar el número particular en donde se da la ubicación del inmueble en una calle dada.
    :param no_interior: Este atributo opcional sirve para expresar información adicional para especificar la ubicación cuando calle y número exterior (noExterior) no resulten suficientes para determinar la ubicación precisa del inmueble.
    :param colonia: Este atributo opcional sirve para precisar la colonia en donde se da la ubicación del inmueble cuando se desea ser más específico en casos de ubicaciones urbanas.
    :param localidad: Atributo opcional que sirve para precisar la ciudad o población donde se da la ubicación del inmueble
    :param referencia: Atributo opcional para expresar una referencia adicional de ubicación del inmueble.
    """
    
    def __init__(
            self,
            tipo_inmueble: str,
            calle: str,
            municipio: str,
            estado: str,
            pais: str,
            codigo_postal: str,
            no_exterior: str = None,
            no_interior: str = None,
            colonia: str = None,
            localidad: str = None,
            referencia: str = None,
    ): 
        super().__init__({
            'TipoInmueble': tipo_inmueble,
            'Calle': calle,
            'Municipio': municipio,
            'Estado': estado,
            'Pais': pais,
            'CodigoPostal': codigo_postal,
            'NoExterior': no_exterior,
            'NoInterior': no_interior,
            'Colonia': colonia,
            'Localidad': localidad,
            'Referencia': referencia,
        })
        

class NotariosPublicos(CFDI):
    """
    Complemento al Comprobante Fiscal Digital a través de Internet (CFDI) para el manejo de la enajenación de bienes inmuebles o servidumbres de paso con indemnización o contraprestación en una sola exhibición.
    
    :param desc_inmuebles: Nodo que contiene las descripciones del inmueble o inmuebles objeto del acto otorgado.
    :param datos_operacion: Nodo para definir los detalles de la operación.
    :param datos_notario:
    :param datos_enajenante: Nodo para capturar los datos del enajenante o enajenantes, o en el caso de servidumbres de paso del propietario o poseedores o propietarios o poseedores del predio sirviente.
    :param datos_adquiriente: Nodo para capturar los datos del adquiriente, adquirientes o propietario, o propietarios o poseedores, en caso de servidumbres de paso.
    """
    
    tag = '{http://www.sat.gob.mx/notariospublicos}NotariosPublicos'
    version = '1.0'
    
    def __init__(
            self,
            desc_inmuebles: DescInmueble | dict | Sequence[DescInmueble | dict],
            datos_operacion: DatosOperacion | dict,
            datos_notario: DatosNotario | dict,
            datos_enajenante: DatosEnajenante | dict,
            datos_adquiriente: DatosAdquiriente | dict,
    ): 
        super().__init__({
            'Version': self.version,
            'DescInmuebles': desc_inmuebles,
            'DatosOperacion': datos_operacion,
            'DatosNotario': datos_notario,
            'DatosEnajenante': datos_enajenante,
            'DatosAdquiriente': datos_adquiriente,
        })
        




"""cce11 http://www.sat.gob.mx/ComercioExterior11"""
from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ... import CFDI, XElement, ScalarMap


class DescripcionesEspecificas(ScalarMap):
    """
    Nodo opcional que indica la lista de descripciones específicas de la mercancía. Una mercancía puede tener más de una descripción específica.
    
    :param marca: Atributo requerido que indica la marca de la mercancía.
    :param modelo: Atributo opcional que indica el modelo de la mercancía.
    :param sub_modelo: Atributo opcional que indica el submodelo de la mercancía.
    :param numero_serie: Atributo opcional que indica el número de serie de la mercancía.
    """
    
    def __init__(
            self,
            marca: str,
            modelo: str = None,
            sub_modelo: str = None,
            numero_serie: str = None,
    ): 
        super().__init__({
            'Marca': marca,
            'Modelo': modelo,
            'SubModelo': sub_modelo,
            'NumeroSerie': numero_serie,
        })
        

class Mercancia(ScalarMap):
    """
    Nodo requerido para capturar la información de la declaración de cada mercancía exportada.
    
    :param no_identificacion: Atributo requerido que sirve para expresar el número de parte, la clave de identificación que asigna la empresa o el número de serie de la mercancía exportada.
    :param valor_dolares: Atributo requerido que indica el valor total en dólares de Estados Unidos (USD).
    :param fraccion_arancelaria: Atributo condicional que sirve para expresar la clave de la fracción arancelaria correspondiente a la descripción de la mercancía exportada, este dato se vuelve requerido cuando se cuente con él o se esté obligado legalmente a contar con él.Debe ser conforme con el catálogo c_FraccionArancelaria publicado en el portal del SAT en internet.
    :param cantidad_aduana: Atributo opcional para precisar la cantidad de bienes en la aduana conforme a la UnidadAduana cuando en el nodo Comprobante:Conceptos:Concepto se hubiera registrado información comercial.
    :param unidad_aduana: Atributo condicional para precisar la clave de la unidad de medida aplicable para la cantidad expresada en la mercancía en la aduana, conforme con el catálogo c_UnidadAduana publicado en el portal del SAT en internet.
    :param valor_unitario_aduana: Atributo condicional para precisar el valor o precio unitario del bien en la aduana. Se expresa en dólares de Estados Unidos (USD), el cual puede estar registrado hasta centésimas.
    :param descripciones_especificas: Nodo opcional que indica la lista de descripciones específicas de la mercancía. Una mercancía puede tener más de una descripción específica.
    """
    
    def __init__(
            self,
            no_identificacion: str,
            valor_dolares: Decimal | int,
            fraccion_arancelaria: str = None,
            cantidad_aduana: Decimal | int = None,
            unidad_aduana: str = None,
            valor_unitario_aduana: Decimal | int = None,
            descripciones_especificas: DescripcionesEspecificas | dict | Sequence[DescripcionesEspecificas | dict] = None,
    ): 
        super().__init__({
            'NoIdentificacion': no_identificacion,
            'ValorDolares': valor_dolares,
            'FraccionArancelaria': fraccion_arancelaria,
            'CantidadAduana': cantidad_aduana,
            'UnidadAduana': unidad_aduana,
            'ValorUnitarioAduana': valor_unitario_aduana,
            'DescripcionesEspecificas': descripciones_especificas,
        })
        

class Domicilio(ScalarMap):
    """
    Nodo requerido para expresar el domicilio del destinatario de la mercancía.
    
    :param calle: Atributo requerido sirve para precisar la calle en que está ubicado el domicilio del destinatario de la mercancía.
    :param estado: Atributo requerido para señalar el estado, entidad, región, comunidad u otra figura análoga en donde se encuentra ubicado el domicilio del destinatario de la mercancía. El catálogo se publica en el portal del SAT en internet y es conforme con la especificación ISO 3166-2.
    :param pais: Atributo requerido que sirve para precisar la clave del país donde se encuentra ubicado el destinatario de la mercancía, conforme con el catálogo c_Pais publicado en el portal del SAT en internet que está basado en la especificación ISO 3166-1.
    :param codigo_postal: Atributo requerido que sirve para asentar el código postal (PO, BOX) en donde se encuentra ubicado el domicilio del destinatario de la mercancía.
    :param numero_exterior: Atributo opcional sirve para expresar el número exterior en donde se ubica el domicilio del destinatario de la mercancía.
    :param numero_interior: Atributo opcional sirve para expresar el número interior, en caso de existir, en donde se ubica el domicilio del destinatario de la mercancía.
    :param colonia: Atributo opcional sirve para expresar la colonia o dato análogo en donde se ubica el domicilio del destinatario de la mercancía.
    :param localidad: Atributo opcional que sirve para precisar la ciudad, población, distrito u otro análogo en donde se ubica el domicilio del destinatario de la mercancía.
    :param referencia: Atributo opcional para expresar una referencia geográfica adicional que permita una fácil o precisa ubicación del domicilio del destinatario de la mercancía, por ejemplo las coordenadas GPS.
    :param municipio: Atributo opcional que sirve para precisar el municipio, delegación, condado u otro análogo en donde se encuentra ubicado el destinatario de la mercancía.
    """
    
    def __init__(
            self,
            calle: str,
            estado: str,
            pais: str,
            codigo_postal: str,
            numero_exterior: str = None,
            numero_interior: str = None,
            colonia: str = None,
            localidad: str = None,
            referencia: str = None,
            municipio: str = None,
    ): 
        super().__init__({
            'Calle': calle,
            'Estado': estado,
            'Pais': pais,
            'CodigoPostal': codigo_postal,
            'NumeroExterior': numero_exterior,
            'NumeroInterior': numero_interior,
            'Colonia': colonia,
            'Localidad': localidad,
            'Referencia': referencia,
            'Municipio': municipio,
        })
        

class Destinatario(ScalarMap):
    """
    Nodo opcional para capturar los datos del destinatario de la mercancía cuando éste sea distinto del receptor del CFDI.
    
    :param domicilio: Nodo requerido para expresar el domicilio del destinatario de la mercancía.
    :param num_reg_id_trib: Atributo opcional para incorporar el número de identificación o registro fiscal del país de residencia para efectos fiscales del destinatario de la mercancía exportada.
    :param nombre: Atributo opcional para expresar el nombre completo, denominación o razón social del destinatario de la mercancía exportada.
    """
    
    def __init__(
            self,
            domicilio: Domicilio | dict | Sequence[Domicilio | dict],
            num_reg_id_trib: str = None,
            nombre: str = None,
    ): 
        super().__init__({
            'Domicilio': domicilio,
            'NumRegIdTrib': num_reg_id_trib,
            'Nombre': nombre,
        })
        

class Receptor(ScalarMap):
    """
    Nodo condicional para capturar los datos complementarios del receptor del CFDI.
    
    :param num_reg_id_trib: Atributo condicional para incorporar el número de identificación o registro fiscal del país de residencia para efectos fiscales del receptor del CFDI.
    :param domicilio: Nodo requerido para expresar el domicilio del receptor del comprobante.
    """
    
    def __init__(
            self,
            num_reg_id_trib: str = None,
            domicilio: Domicilio | dict = None,
    ): 
        super().__init__({
            'NumRegIdTrib': num_reg_id_trib,
            'Domicilio': domicilio,
        })
        

class Propietario(ScalarMap):
    """
    Nodo condicional para capturar los datos del o los propietarios de la mercancía que se traslada y ésta no sea objeto de enajenación o siéndolo sea a título gratuito, cuando el emisor del CFDI es un tercero.
    
    :param num_reg_id_trib: Atributo requerido para incorporar el número de identificación o registro fiscal del país de residencia para efectos fiscales del propietario de la mercancía trasladada.
    :param residencia_fiscal: Atributo requerido para registrar la clave del país de residencia para efectos fiscales del propietario de la mercancía, conforme con el catálogo c_Pais publicado en el portal del SAT en internet que está basado en la especificación ISO 3166-1.
    """
    
    def __init__(
            self,
            num_reg_id_trib: str,
            residencia_fiscal: str,
    ): 
        super().__init__({
            'NumRegIdTrib': num_reg_id_trib,
            'ResidenciaFiscal': residencia_fiscal,
        })
        

class Emisor(ScalarMap):
    """
    Nodo opcional para capturar los datos complementarios del emisor del comprobante.
    
    :param curp: Atributo condicional para expresar la CURP del emisor del CFDI cuando es una persona física.
    :param domicilio: Nodo condicional para expresar el domicilio del emisor del comprobante, es requerido cuando el complemento se registre en un CFDI 4.0.
    """
    
    def __init__(
            self,
            curp: str = None,
            domicilio: Domicilio | dict = None,
    ): 
        super().__init__({
            'Curp': curp,
            'Domicilio': domicilio,
        })
        

class ComercioExterior(CFDI):
    """
    Complemento para incorporar la información en el caso de Exportación de Mercancías en definitiva.
    
    :param tipo_operacion: Atributo requerido que indica la clave del tipo de operación de Comercio Exterior que se realiza, conforme con el catálogo c_TipoOperacion publicado en el portal del SAT en internet.
    :param motivo_traslado: Atributo condicional que indica la clave del motivo por el cual en la exportación definitiva de mercancías con clave de pedimento A1, éstas no son objeto de enajenación o siéndolo sean a título gratuito, desde el domicilio del emisor hacia el domicilio del receptor o del destinatario. La clave del motivo es conforme con el catálogo c_MotivoTraslado publicado en el portal del SAT en internet.
    :param clave_de_pedimento: Atributo condicional que indica la clave de pedimento que se haya declarado conforme con el catálogo c_ClavePedimento publicado en el portal del SAT en internet.
    :param certificado_origen: Atributo condicional derivado de la excepción de certificados de Origen de los Tratados de Libre Comercio que ha celebrado México con diversos países. 0 = No Funge como certificado de origen 1 = Funge como certificado de origen.
    :param num_certificado_origen: Atributo condicional para expresar el folio del certificado de origen o el folio fiscal del CFDI con el que se pagó la expedición del certificado de origen.
    :param numero_exportador_confiable: Atributo condicional que indica el número de exportador confiable, conforme al artículo 22 del Anexo 1 del Tratado de Libre Comercio con la Asociación Europea y a la Decisión de la Comunidad Europea.
    :param incoterm: Atributo condicional que indica la clave del INCOTERM aplicable a la factura, conforme con el catálogo c_INCOTERM publicado en el portal del SAT en internet.
    :param subdivision: Atributo condicional que indica si la factura tiene o no subdivisión. Valores posibles: 0 - no tiene subdivisión,1 - si tiene subdivisión.
    :param observaciones: Atributo opcional en caso de ingresar alguna información adicional, como alguna leyenda que debe incluir en el CFDI.
    :param tipo_cambio_usd: Atributo condicional que indica el número de pesos mexicanos que equivalen a un dólar de Estados Unidos, de acuerdo al artículo 20 del Código Fiscal de la Federación.
    :param total_usd: Atributo condicional que indica el importe total del comprobante en dólares de Estados Unidos.
    :param emisor: Nodo opcional para capturar los datos complementarios del emisor del comprobante.
    :param propietario: Nodo condicional para capturar los datos del o los propietarios de la mercancía que se traslada y ésta no sea objeto de enajenación o siéndolo sea a título gratuito, cuando el emisor del CFDI es un tercero.
    :param receptor: Nodo condicional para capturar los datos complementarios del receptor del CFDI.
    :param destinatario: Nodo opcional para capturar los datos del destinatario de la mercancía cuando éste sea distinto del receptor del CFDI.
    :param mercancias: Nodo condicional para capturar la información de la declaración de las mercancías exportadas.
    """
    
    tag = '{http://www.sat.gob.mx/ComercioExterior11}ComercioExterior'
    version = '1.1'
    
    def __init__(
            self,
            tipo_operacion: str,
            motivo_traslado: str = None,
            clave_de_pedimento: str = None,
            certificado_origen: int = None,
            num_certificado_origen: str = None,
            numero_exportador_confiable: str = None,
            incoterm: str = None,
            subdivision: int = None,
            observaciones: str = None,
            tipo_cambio_usd: Decimal | int = None,
            total_usd: Decimal | int = None,
            emisor: Emisor | dict = None,
            propietario: Propietario | dict | Sequence[Propietario | dict] = None,
            receptor: Receptor | dict = None,
            destinatario: Destinatario | dict | Sequence[Destinatario | dict] = None,
            mercancias: Mercancia | dict | Sequence[Mercancia | dict] = None,
    ): 
        super().__init__({
            'Version': self.version,
            'TipoOperacion': tipo_operacion,
            'MotivoTraslado': motivo_traslado,
            'ClaveDePedimento': clave_de_pedimento,
            'CertificadoOrigen': certificado_origen,
            'NumCertificadoOrigen': num_certificado_origen,
            'NumeroExportadorConfiable': numero_exportador_confiable,
            'Incoterm': incoterm,
            'Subdivision': subdivision,
            'Observaciones': observaciones,
            'TipoCambioUSD': tipo_cambio_usd,
            'TotalUSD': total_usd,
            'Emisor': emisor,
            'Propietario': propietario,
            'Receptor': receptor,
            'Destinatario': destinatario,
            'Mercancias': mercancias,
        })
        




import base64
import hashlib
import io
import os.path
from datetime import datetime
from random import randbytes

from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from cryptography.hazmat.primitives.serialization import pkcs7

from .. import Signer, Certificate, RFC, RFCType
from ..ans1e import *
from ..models.certificate import CertificateType
from ..zip import zip_create, ZipData

ENCODING = 'windows-1252'


def _p_join(dirname, filename):
    if dirname is None:
        return filename
    os.makedirs(dirname, exist_ok=True)
    return os.path.join(dirname, filename)


class Certifica:
    def __init__(self, signer: Signer):
        self.signer = signer
        if self.signer.type != CertificateType.Fiel:
            raise ValueError("'signer' no es Fiel")

    def generacion_fiel(self, password: str, rfc: str, curp: str, correo: str = None, dirname: str = None):
        tmf = datetime.now().strftime('%Y%m%d_%H%M%S')
        key_filename = f"Claveprivada_FIEL_{self.signer.rfc}_{tmf}.key"
        private_key = self._generate_key(_p_join(dirname, key_filename), password)

        ren_filename = f"Renovacion_FIEL_{self.signer.rfc}_{tmf}.req"
        data = _create_generacion_certificate_signing_request(private_key=private_key, rfc=rfc, curp=curp, correo=correo)
        with open(_p_join(dirname, ren_filename), 'wb') as f:
            f.write(data)

    def renovacion_fiel(self, password: str, correo: str = None, dirname: str = None):
        tmf = datetime.now().strftime('%Y%m%d_%H%M%S')
        key_filename = f"Claveprivada_FIEL_{self.signer.rfc}_{tmf}.key"
        private_key = self._generate_key(_p_join(dirname, key_filename), password)

        ren_filename = f"Renovacion_FIEL_{self.signer.rfc}_{tmf}.ren"
        data = _create_renovation_certificate_signing_request(signer=self.signer, private_key=private_key, correo=correo)
        pck = self._pkcs7_package(data)
        with open(_p_join(dirname, ren_filename), 'wb') as f:
            f.write(pck)

    def renovacion_fiel_moral(self, password: str, rfc: str, correo: str, dirname: str = None):
        rfc = RFC(rfc)
        if rfc.type != RFCType.MORAL:
            raise ValueError("RFC no es Moral")
        rfc = str(rfc)

        tmf = datetime.now().strftime('%Y%m%d_%H%M%S')
        key_filename = f"Claveprivada_FIEL_{rfc}_{tmf}.key"
        private_key = self._generate_key(_p_join(dirname, key_filename), password)

        ren_filename = f"Renovacion_FIEL_{rfc}_{tmf}.ren"
        data = _create_renovation_moral_certificate_signing_request(
            signer=self.signer,
            private_key=private_key,
            rfc=rfc,
            correo=correo
        )
        pck = self._pkcs7_package(data)
        with open(_p_join(dirname, ren_filename), 'wb') as f:
            f.write(pck)

    def solicitud_certificado(self, sucursal: str, password: str, dirname: str = None):
        tmf = datetime.now().strftime('%Y%m%d_%H%M%S')
        key_filename = f"CSD_{sucursal.replace(' ', '_')}_{self.signer.rfc}_{tmf}.key"
        private_key = self._generate_key(_p_join(dirname, key_filename), password)

        sdg_filename = f"CSD_{self.signer.rfc}_{tmf}.sdg"
        data = _create_certificate_signing_request_zip(self.signer, private_key, sucursal)
        pck = self._pkcs7_package(data)
        with open(_p_join(dirname, sdg_filename), 'wb') as f:
            f.write(pck)

    @staticmethod
    def _generate_key(key_filename, password):
        private_key = rsa.generate_private_key(
            public_exponent=65537,
            key_size=2048,
        )

        p_key = private_key.private_bytes(
            encoding=serialization.Encoding.DER,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.BestAvailableEncryption(password.encode())
        )

        with open(key_filename, 'wb') as f:
            f.write(p_key)

        return private_key

    def _pkcs7_package(self, data):
        cert = self.signer.certificate.to_cryptography()
        key = self.signer.key
        options = [pkcs7.PKCS7Options.NoCapabilities, pkcs7.PKCS7Options.Binary]

        return pkcs7.PKCS7SignatureBuilder().set_data(data) \
            .add_signer(cert, key, hashes.SHA1()) \
            .sign(serialization.Encoding.DER, options)


def _create_certificate_signing_request_zip(signer: Signer, private_key: RSAPrivateKey, sucursal: str):
    res = _create_certificate_signing_request(signer=signer, private_key=private_key, sucursal=sucursal)

    tmf = datetime.now().strftime('%Y%m%d_%H%M%S')
    tmp_filename = f"CSD_{sucursal.replace(' ', '_')}_{signer.rfc}_{tmf}s.req"

    with io.BytesIO() as b:
        zip_create(b, files=[ZipData(tmp_filename, lambda s: s.write(res))])
        return b.getvalue()


def _create_generacion_certificate_signing_request(private_key: RSAPrivateKey, rfc: str, curp: str, correo: str):
    code = _calculate_code_random()

    subject = {
        '2.5.4.45': (rfc.encode(ENCODING), Numbers.PrintableString),
        '2.5.4.5': (curp.encode(ENCODING), Numbers.PrintableString),
        '1.2.840.113549.1.9.1': (correo, Numbers.IA5String)
    }

    return _certificate_request(private_key, subject, code)


def _create_renovation_certificate_signing_request(signer: Signer, private_key: RSAPrivateKey, correo):
    code = _calculate_code_random()
    cer = signer.certificate
    subject_at = {
        k: v for k, v in cer.get_subject().get_components()
    }

    subject = {
        '2.5.4.45': (subject_at[b'x500UniqueIdentifier'], Numbers.PrintableString),
        '2.5.4.5': (subject_at[b'serialNumber'], Numbers.PrintableString),
        '1.2.840.113549.1.9.1': (correo or subject_at[b'emailAddress'], Numbers.IA5String)
    }
    return _certificate_request(private_key, subject, code)


def _create_renovation_moral_certificate_signing_request(signer: Signer, private_key: RSAPrivateKey, rfc, correo):
    code = _calculate_code_random()
    cer = signer.certificate
    subject_at = {
        k: v for k, v in cer.get_subject().get_components()
    }

    print(signer.rfc)
    print(subject_at[b'x500UniqueIdentifier'])
    print(subject_at[b'serialNumber'])

    subject = {
        '2.5.4.45': (rfc.encode(ENCODING) + b' / ' + subject_at[b'x500UniqueIdentifier'], Numbers.PrintableString),
        '2.5.4.5': (b' / ' + subject_at[b'serialNumber'], Numbers.PrintableString),
        '1.2.840.113549.1.9.1': (correo, Numbers.IA5String)
    }
    return _certificate_request(private_key, subject, code)


def _create_certificate_signing_request(signer: Signer, private_key: RSAPrivateKey, sucursal: str):
    code = _calculate_code(signer)
    cer = signer.certificate
    subject_at = {
        k: v for k, v in cer.get_subject().get_components()
    }

    subject = {
        '2.5.4.45': (subject_at[b'x500UniqueIdentifier'], Numbers.PrintableString),
        '2.5.4.5': (subject_at[b'serialNumber'], Numbers.PrintableString)
    }
    if RFC(signer.rfc).type == RFCType.MORAL:
        subject['2.5.4.10'] = (subject_at[b'O'].decode(ENCODING), None)
    else:
        subject['2.5.4.3'] = (subject_at[b'CN'].decode(ENCODING), None)
    subject['2.5.4.11'] = (sucursal, None)

    return _certificate_request(private_key, subject, code)


def _certificate_request(private_key: RSAPrivateKey, subject: dict, code: bytes):
    public_key_bytes = private_key.public_key().public_bytes(encoding=serialization.Encoding.DER, format=serialization.PublicFormat.PKCS1)

    e = Ans1Encoder()
    with e.seq():
        e(0)
        with e.seq():
            for k, (v, nr) in subject.items():
                with e.set():
                    with e.seq():
                        e.oid(k)
                        e(v, nr=nr)
        with e.seq():
            with e.seq():
                e.oid('1.2.840.113549.1.1.1')
                e()
            e(public_key_bytes, nr=Numbers.BitString)
        with e.enter(nr=0, cls=Classes.Context):
            with e.seq():
                e.oid('1.2.840.113549.1.9.7')
                with e.set():
                    e(code, nr=Numbers.PrintableString)
    cert_request_bytes = e.output()

    e = Ans1Encoder()
    with e.seq():
        e.write(cert_request_bytes)
        with e.seq():
            e.oid('1.2.840.113549.1.1.5')
            e()
        e(
            private_key.sign(data=cert_request_bytes, padding=padding.PKCS1v15(), algorithm=hashes.SHA1()),
            nr=Numbers.BitString
        )
    return e.output()


def _calculate_code(certificate: Certificate):
    ui = next(v for k, v in certificate.certificate.get_subject().get_components() if k == b'x500UniqueIdentifier')

    def digest(value):
        m = hashlib.sha1()
        m.update(value)
        return base64.b64encode(m.digest())

    return digest(ui + digest(ui + ui))


def _calculate_code_random():
    return base64.b64encode(randbytes(20))
